#include <stdio.h>
#include <stdlib.h>

#include "tinecdi.h"
#include "dbtools.h"
#include "cdi.h"	// interface to user program
#include "cdiCanLib.h"

#ifndef CDI_CAN_LINUX
# include "windows.h"
#define snprintf _snprintf

#else

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>   // exit
#include <signal.h>
#include <string.h>
#include <stdlib.h>   // strtoul
#include <fcntl.h>    // O_RDWR
#include <ctype.h>
/*
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <memory.h>
#include <fcntl.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <setjmp.h>

#include <math.h>
#include <stropts.h>
#include <semaphore.h>
*/
#endif

char *canMsgTypeString[] = {"SDO: ","PDO1: ","PDO2: ","NMT: ","EMERGENCY: ","NotDefined: "};

char *scanCommand[] = {"getcommands","getlines","getrcvdmsg","getmodules","readsdo","writesdo",
      "readpdo1","writepdo1","readpdo2","writepdo2","writenmt", "readnmt",
      "startremote","stopremote","enterpreoperation","resetnode","resetcommunication"};
enum scommand {getcommands, getlines, getrcvdmsg, getmodules, readsdo, writesdo,
      readpdo1, writepdo1, readpdo2, writepdo2, writenmt, readnmt,
      startremote, stopremote, enterpreoperation, resetnode, resetcommunication, nCommands};

#define WAIT_CANMSG_MS  600000000      // 600ms
#define TXT_TIMEOUT     200           // ms
char *scanBuffer;
unsigned int scanBufferSize;
int scanFlag = 0;
int scanNodId = 0;

unsigned int numStoredMsg[MAX_CAN_IDS];
sem_t scan_SemId;

#ifdef CDI_CAN_LINUX
__attribute__((constructor)) void cdiCanMain(void)
{
  int cc = 0;

  if ((cc=cdiRegisterBusFilter(CAN_LIB_NAME,filterCanbus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusConfiguration(CAN_LIB_NAME,configCanCard)) != 0) goto err;
  if ((cc=cdiRegisterBusInitialization(CAN_LIB_NAME,initCanLineForCDI)) != 0 ) goto err;
  if ((cc=cdiRegisterBusHandler(CAN_LIB_NAME,execCanbus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusCleanup(CAN_LIB_NAME,terminateCanLine)) != 0 ) goto err;
  if ((cc=cdiRegisterBusScanner(CAN_LIB_NAME,canScanFcn)) != 0 ) goto err;

err:
  if (cc)
  {
    cdilog("%s : failure in bus registration CAN",erlst[cc]);
  }
  return;
}
#endif

int NGdebug = 0;
int canMsgPrint = 0;
#define MAX_BLEN 256

CAN_LINE canLine[MAX_NUMBER_CAN_LINES] = {
	{-1,0}, {-1,0}, {-1,0}, {-1,0}
//#ifdef CDI_CAN_LINUX
//	{-1,0}, {-1,0}, {-1,0}, {-1,0}
//#else
//	{-1,NULL}, {-1,NULL}, {-1,NULL}, {-1,NULL}
//#endif
};

CDI_CAN_EXPORT int canScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256], canData[8];
  int i, j, cnt, k, max, ret, slen = 0;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;
  struct timespec wait_time;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0 ) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0); //atoi(c);

  if (j == getcommands)
  {
    snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }  
  
  if ( j == getlines )
  {
    snprintf(result,resultBufferSize,"CAN BUS lines configured:\n");
    for (i=0, cnt=0; i<MAX_NUMBER_CAN_LINES; i++)
    {
      if ( canLine[i].cdiLine < 0 ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
       snprintf(&result[slen],resultBufferSize-slen,"%d",i+1);
      cnt++;
    }
    return 0;
  }
  
  line = data[0];
  if ( nData < 1 || line < 1 || line > MAX_NUMBER_CAN_LINES || canLine[line-1].cdiLine < 0) return no_such_line;
  scanBufferSize = resultBufferSize;
  scanBuffer = result;

  if ( j == getrcvdmsg )
  {
    crate = ( nData > 1 && data[1] > 0 && data[1] < MAX_CAN_IDS ) ? data[1] : 0;
    if ( crate ) { i = crate; max = crate + 1; }
    else { i = 0; max = MAX_CAN_IDS; };
    snprintf(result,resultBufferSize,"received CAN messages on line %d:\n",line);
    for (; i<max; i++)
    {
      for (k=0, cnt=0; k<numTypeMsg; k++)
      {
        if ( canLine[line-1].numRevMsg[k][i] == 0 ) continue;
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
        if ( cnt > 0 )
        {
          if ((slen = strlen(result)) > resultBufferSize - 1) break;
          snprintf(&result[slen],resultBufferSize-slen,", ");
        }
        else
        {
          if ((slen = strlen(result)) > resultBufferSize - 15) break;
          snprintf(&result[slen],resultBufferSize-slen,"NodId:%d received:\n",i);
        }
        if ((slen = strlen(result)) > resultBufferSize - 10) break;
        snprintf(&result[slen],resultBufferSize-slen,"%s",canMsgTypeString[k]);
        if ((slen = strlen(result)) > resultBufferSize - 4) break;
        snprintf(&result[slen],resultBufferSize-slen,"%d",canLine[line-1].numRevMsg[k][i]);
        cnt++;
      }
      if ( cnt > 0 )
      {
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
        snprintf(&result[slen],resultBufferSize-slen,"\n");
      }
    }
    return 0;
  }
  if ( j >= startremote && j <= resetcommunication )
  {
    crate = ( nData > 1 && data[1] > 0 && data[1] < MAX_CAN_IDS ) ? data[1] : 0;
    snprintf(result,resultBufferSize,"%s on line %d for Node: %d\n",scanCommand[j], line, crate);

    scanFlag = j;
    scanNodId = 0;

    if ( j == startremote ) wData = NMT_START_REMOTE;
    else if ( j == stopremote ) wData = NMT_STOP_REMOTE;
    else if ( j == enterpreoperation ) wData = NMT_ENTER_PREOPERATION;
    else if ( j == resetnode ) wData = NMT_RESET_NODE;
    else wData = NMT_RESET_COMMUNICATION;

  	if ( ret = canMsgSend(line, crate, NMT_COMMANDS, 1, (char *)&wData, 0, 0, 0)  )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
    }
    else
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = WAIT_CANMSG_MS; 
      nanosleep( &wait_time, NULL );
    }
    scanFlag = 0;
    return 0;
  }
  if ( j == getmodules )  // request pdos
  {
    snprintf(result,resultBufferSize,"NodIDs connected on line %d:\n",line);
    for (i=0; i<MAX_CAN_IDS; i++) numStoredMsg[i] = canLine[line-1].numRevMsg[pdo1Msg][i] + canLine[line-1].numRevMsg[pdo2Msg][i];

//    shared = value = 0;
//    if ((retSem=cdiGetSemaphore(&scan_SemId,shared,value)) != 0)
//    {
//      if ((slen = strlen(result)) < resultBufferSize - 20)
//        snprintf(&result[slen],resultBufferSize-slen,"GetSemaphore Error!");
//      return 0;
//    }
  	
    // start remote to use pdos
    crate = 0;
    wData = NMT_START_REMOTE;

  	if ( ret = canMsgSend(line, crate, NMT_COMMANDS, 1, (char *)&wData, 0, 0, 0) )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
      return 0;
    }
    // request pdo1 and pdo2
  	for (  crate=1; crate<MAX_CAN_IDS; crate++ )
    {
    	ret = canMsgSend(line, crate, PDO_DI_READ, 0, NULL, 0, 0, 0);
      ret = canMsgSend(line, crate, PDO_AI_READ, 0, NULL, 0, 0, 0);
    	if ( ret = canMsgSend(line, crate, NMT_COMMANDS, 1, (char *)&wData, 0, 0, 0) )
      {
        if ((slen = strlen(result)) < resultBufferSize - 20)
          snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
        return 0;
      }
    }

    wait_time.tv_sec = 0;
    wait_time.tv_nsec = WAIT_CANMSG_MS; 
    nanosleep( &wait_time, NULL );
//    cdiWaitSemaphore(&scan_SemId,1);
//printf("wait finished!\n");
    for (i=0, cnt=0; i<MAX_CAN_IDS; i++)
    {
      if ( canLine[line-1].numRevMsg[pdo1Msg][i] + canLine[line-1].numRevMsg[pdo2Msg][i] <= numStoredMsg[i] ) continue;
      if ( cnt > 0 )
      {
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
        snprintf(&result[slen],resultBufferSize-slen,",");
      }
      if ((slen = strlen(result)) > resultBufferSize - 4) break;
      snprintf(&result[slen],resultBufferSize-slen,"%d",i);
      cnt++;
    }
//    cdiDestroySemaphore(&scan_SemId);
    return 0;
  }

  if ( nData < 2 || data[1] < 1 || data[1] >= MAX_CAN_IDS ) return argument_list_error;
  crate = data[1];
  if ( j == readpdo1 || j == readpdo2 || j == readnmt)  // read pdo & nmt messages
  {
    snprintf(result,resultBufferSize,"%s on line %d for Node: %d\n",scanCommand[j], line, crate);
    scanFlag = j;
    scanNodId = crate;
    if ( j != readnmt && (ret = canMsgSend(line, crate, (short)(j == readpdo1 ? PDO_DI_READ : PDO_AI_READ), 0, NULL, 0, 0, 0)) )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
    }
    else
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = WAIT_CANMSG_MS; 
      nanosleep( &wait_time, NULL );
    }
    scanFlag = 0;
    scanNodId = 0;
    return 0;
  }

  if ( nData < 3 ) return argument_list_error;
  if ( j == writepdo1 || j == writepdo2 || j == writenmt)  // write pdo and nmt messages
  {
    snprintf(result,resultBufferSize,"%s on line %d for Node: %d\n",scanCommand[j], line, crate);
    for (i=2; i<nData && i<10; i++ ) canData[i-2] = (char)data[i];
    rData = i - 2;

    if ( j == writepdo1 ) subaddr = PDO_TRANSMIT;  
    else if ( j == writepdo2 ) subaddr = PDO2_TRANSMIT;  
    else subaddr = NMT_COMMANDS;

    scanFlag = j;
    scanNodId = crate;
    if ( ret = canMsgSend(line, crate, subaddr, rData, canData, 0, 0, 1) )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
    }
    else
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = WAIT_CANMSG_MS; 
      nanosleep( &wait_time, NULL );
    }    
    scanFlag = 0;
    scanNodId = 0;
    return 0;
  }

  if ( nData < 4 ) return argument_list_error;
  if ( j == readsdo || j == writesdo)  // read sdo
  {
    for (i=4; i<nData && i<8; i++ ) canData[i-4] = (char)data[i];
    rData = i - 4;
    if ( j == writesdo && rData < 1 ) return argument_list_error;

    scanFlag = j;
    scanNodId = crate;
    if ( ret = canMsgSend(line, crate, SDO_TRANSMIT, rData, canData, (short)data[2], (short)data[3], (short)(j == readsdo ? 0 : 1)) )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
    }
    else
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = WAIT_CANMSG_MS; 
      nanosleep( &wait_time, NULL );
    }
    scanFlag = 0;
    scanNodId = 0;
    return 0;
  }

  return 0;
}

CDI_CAN_EXPORT int filterCanbus(int *parameter, int accessFlag)
{
  if (parameter == NULL) return argument_list_error;
  if (parameter[msgType] < sdoMsg || parameter[msgType] > notdefinedMsg ) return argument_list_error;

  /* only read possible */
  if (accessFlag != cdiReadFlag && (parameter[msgType] == emergencyMsg || parameter[msgType] == notdefinedMsg) ) return argument_list_error;
  
  /* set canAutoReceive flag */
  if (parameter[msgType] == emergencyMsg || parameter[msgType] == notdefinedMsg) parameter[canAutoReceive] = 1;
  else if (parameter[msgType] == nmtMsg && accessFlag != cdiWriteFlag) parameter[canAutoReceive] = 1;

  /* set canOpenId & canOpenReadId */
  if ( parameter[msgType] == sdoMsg ) { parameter[canOpenId] = SDO_TRANSMIT;  parameter[canOpenReadId] = SDO_TRANSMIT;}
  else if ( parameter[msgType] == pdo1Msg ) { parameter[canOpenId] = PDO_TRANSMIT;  parameter[canOpenReadId] = PDO_DI_READ;}
  else if ( parameter[msgType] == pdo2Msg ) { parameter[canOpenId] = PDO2_TRANSMIT;  parameter[canOpenReadId] = PDO_AI_READ;}
  else if ( parameter[msgType] == nmtMsg ) { parameter[canOpenId] = NMT_COMMANDS;  parameter[canOpenReadId] = NMT_ERROS;}
  else if ( parameter[msgType] == emergencyMsg ) { parameter[canOpenId] = NMT_COMMANDS;  parameter[canOpenReadId] = EMERGENCE_MESSAGE;}
  else { parameter[canOpenId] = NMT_COMMANDS;  parameter[canOpenReadId] = notdefinedMsg;} /* any messages */

  return 0;
}

/****************************************************************************/
int sendCdiCanMessage(CdiRequestInfoBlk *pReq, int wFlag)
{ int canRet;

  canRet = canMsgSend((int)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine,
              (short)(pReq->devParameters[canBroadcast] ? 0 : pReq->pDev->cdiAddr),
              (short)pReq->devParameters[wFlag ? canOpenId : canOpenReadId], 
              (short)pReq->outLength, 
              (char *)pReq->pOutData,
              (short)pReq->devParameters[canOpenObj], 
              (short)pReq->devParameters[canOpenSubIndex], 
              (short)wFlag);

  if ( NGdebug )
    cdilog("canMsgSend return code: %d! line: %d\n",canRet, ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine);
  
  if ( canRet < 0 )
  {
    pReq->errorCode = -canRet; /* write failed */
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( pReq->errorCode == hardware_error )
    {
      pReq->deviceError = net_write_error;    // can write error
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest;
  }
  return canRet;
}

/****************************************************************************/
/* send request! input active flag = transRequest or transRequest2nd */
/* after, set active flag to finishRequest, recvRequest, or recvRequest2nd */
/****************************************************************************/
CDI_CAN_EXPORT void execCanbus(CdiRequestInfoBlk *pReq)
{ int wFlag;

  if ( pReq->accessFlag > cdiWriteReadFlag )
  {
    pReq->errorCode = not_implemented;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->active = finishRequest; /* finish request */
    return;
  }
  if ( pReq->active == transRequest ) /* start first transfere */
  {
    pReq->active = recvRequest; /* afterwards goto receive */
    /* auto receive, without request! */
    if ( (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) && pReq->devParameters[canAutoReceive] ) return;
    wFlag = (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag) ? 1 : 0;
  }
  else
  {
    pReq->active = recvRequest2nd; /* goto 2ndReceive */
    if ( pReq->accessFlag == cdiWriteReadFlag && pReq->devParameters[canAutoReceive] ) return;
    wFlag = pReq->accessFlag == cdiReadWriteFlag ? 1 : 0;
  }

  if ( sendCdiCanMessage(pReq, wFlag) < 0 ) return;

  /* without message back */
  if ( wFlag && (pReq->devParameters[canOpenId] == PDO_TRANSMIT || pReq->devParameters[canOpenId] == PDO2_TRANSMIT || pReq->devParameters[canOpenId] == NMT_COMMANDS) )
  {
    // last write
    if (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiReadWriteFlag) pReq->active = finishRequest;
    else
    { /* cdiWriteReadFlag */
      pReq->active = recvRequest2nd; /* goto 2ndReceive */
      if ( pReq->devParameters[canAutoReceive] == 0 ) sendCdiCanMessage(pReq, 0);
    }
  }
//  if ( wFlag && (pReq->devParameters[canOpenId] == PDO_TRANSMIT || pReq->devParameters[canOpenId] == PDO2_TRANSMIT || pReq->devParameters[canOpenId] == NMT_COMMANDS) ) pReq->active = finishRequest;

  if ( NGdebug ) cdilog("active = %d; preq addr: 0x%x\n", pReq->active, pReq);

  return;
}

/****************************************/
/*	can receive thread					*/
/****************************************/
#ifdef CDI_CAN_LINUX
void *GetCanDataThread(void *param)
#else
DWORD WINAPI GetCanDataThread(void *param)
#endif
{ int i;
  unsigned int len, errCode, ret;
  char *pErrStr;
  CAN_LINE *pCan = (CAN_LINE *)param;
  do 
  {
//    getCanTelegram(pCan);
//    for ( pCan->recCnt = 0; pCan->recCnt < pCan->recLength; pCan->recCnt++)
//    {
/*      pCan->rcvPmsg = &pCan->cmsg[pCan->recCnt];
      pCan->retId = (short)pCan->rcvPmsg->id;
      pCan->rtr = (pCan->rcvPmsg->len >> 4) & 0x1;
      pCan->length = pCan->rcvPmsg->len & 0xf;

      memcpy(pCan->retData, pCan->rcvPmsg->data, pCan->length);
      cpReceiveTelegram(pCan);
*/
      readCanTelegram(pCan->lineNumber, pCan->fd, &(pCan->retId), &(pCan->rtr), &(pCan->length), pCan->retData);

      pCan->remoteNode = pCan->retId & 0x7f;
      pCan->remoteCommand = pCan->retId & 0xff80;

      if ( pCan->remoteCommand == SDO_RECEIVE ) pCan->numRevMsg[sdoMsg][pCan->remoteNode]++;
      else if ( pCan->remoteCommand == PDO_DI_READ ) pCan->numRevMsg[pdo1Msg][pCan->remoteNode]++;
      else if ( pCan->remoteCommand == PDO_AI_READ ) pCan->numRevMsg[pdo2Msg][pCan->remoteNode]++;
      else if ( pCan->remoteCommand == NMT_ERROS ) pCan->numRevMsg[nmtMsg][pCan->remoteNode]++;
      else if ( pCan->remoteCommand == EMERGENCE_MESSAGE ) pCan->numRevMsg[emergencyMsg][pCan->remoteNode]++;
      else pCan->numRevMsg[notdefinedMsg][pCan->remoteNode]++;

      if ( scanFlag && (len=strlen(scanBuffer)) < scanBufferSize - 30)
      {
        if ( scanFlag >= startremote && scanFlag <= resetcommunication || (pCan->remoteNode == scanNodId &&
          ( ((scanFlag == writepdo1 || scanFlag == readpdo1) && pCan->remoteCommand == PDO_DI_READ) ||
            ((scanFlag == writenmt || scanFlag == readnmt) && (pCan->remoteCommand == NMT_ERROS || pCan->remoteCommand == EMERGENCE_MESSAGE) ) ||
            ((scanFlag == writesdo || scanFlag == readsdo) && pCan->remoteCommand == SDO_RECEIVE) ||
            ((scanFlag == writepdo2 || scanFlag == readpdo2) && pCan->remoteCommand == PDO_AI_READ) ) ) )
        {
          snprintf(&scanBuffer[len],scanBufferSize-len,"Node: %d, MsgID: 0x%x, rtr: %d, Length: %d,",pCan->remoteNode, pCan->remoteCommand, pCan->rtr, pCan->length);
          for(i=0; i<pCan->length; i++)
          {
            if ( (len = strlen(scanBuffer)) < scanBufferSize - 4)
              snprintf(&scanBuffer[len],scanBufferSize-len,"%2x ", ((int)pCan->retData[i] & 0xff));
          }
          if ((scanFlag == writesdo || scanFlag == readsdo) && pCan->remoteCommand == SDO_RECEIVE && (pCan->retData[0] & 0xf0) == 0x80 )// sdo abort Domain Transfer
          {
            errCode = *(unsigned int *)&pCan->retData[4];
            if ( errCode == 0x05040001 ) pErrStr = "Error: illegal Function";
            else if ( errCode == 0x06010002 ) pErrStr = "Error: Access unsupported(Read Only)";
            else if ( errCode == 0x06020000 ) pErrStr = "Error: Object(index) nonexistent";
            else if ( errCode == 0x06070010 ) pErrStr = "Error: Data Type(length) Conflict";
            else if ( errCode == 0x06090011 ) pErrStr = "Error: Subindex nonexistent";
            else pErrStr = "Error: SDO error not defined";
            if ( (len = strlen(scanBuffer)) < scanBufferSize - 20) snprintf(&scanBuffer[len],scanBufferSize-len,"%s",pErrStr);
          }
          if ( (len = strlen(scanBuffer)) < scanBufferSize - 2) snprintf(&scanBuffer[len],scanBufferSize-len,"\n");
        }
      }

      if ( pCan->remoteCommand != NMT_ERROS )
      {
        if ( canMsgPrint )
        {
          printf("Master ID %x, rtr: %d; length: %d : ", pCan->retId, pCan->rtr, pCan->length);
          for(i=0; i<pCan->length; i++) printf("%2x ", ((int)pCan->retData[i] & 0xff));
          printf("\n");
        }
        if ( pCan->receiveCanMessage ) pCan->receiveCanMessage(pCan->cdiLine, pCan->retData, pCan->length, pCan->retId, pCan->rtr, pCan->remoteNode, pCan->remoteCommand);
      }
      else
      {
        /* guarding or heartbeat message: powerOn or preOperation mode */
        if ( (pCan->retData[0] == 0x0 || pCan->retData[0] == 0x7f) && pCan->length == 1 )
        {
          cdilog("Start Operation on Node ID 0x%x, line number: %d\n", pCan->remoteNode, pCan->lineNumber + 1);
          i = NMT_START_REMOTE;
        	if ( ret = canMsgSend(pCan->lineNumber + 1, 0, NMT_COMMANDS, 1, (char *)&i, 0, 0, 0)  )
          {
            cdilog("Start Operation: canMsgSend error(0x%x) on Node ID 0x%x, line number: %d\n", ret, pCan->remoteNode, pCan->lineNumber + 1);
          }
        }
      }
//    }
  } while(1);
}

/****************************************/
/*	CAN send API						*/
/*	return -1: line number wrong		*/
/*	return -2: not init yet				*/
/*	return =0: send error				*/
/****************************************/
int canMsgSend(int line, short nodeId, short canOpenId, short length, char *data, short ObjNum, short subIndex, short wFlag)
{ CAN_LINE *pCan;
  int rtr, ret;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd == 0 ) return -file_not_open;
//  if ( pCan->fd < 0 ) return -file_not_open;

  if ( nodeId < 0 ) nodeId = 0;

//  printf("***canMsgSend ==> line: %d; nodeID: %d; CanOPENID: %d\n",line, nodeId, canOpenId);

  switch (canOpenId)
  {
  	case NMT_COMMANDS:
	    canOpenId = NMT_COMMANDS;
	    rtr = 0;
	    length = 2;
	    pCan->sendData[0] = data[0];	// only one byte data for EXEC_REMOTE
	    pCan->sendData[1] = (unsigned char)nodeId;
	    break;
	  case SDO_TRANSMIT:
	    pCan->sendData[0] = wFlag ? SDO_WRITE : SDO_READ;
	    *(short *)&pCan->sendData[1] = ObjNum;	// object 6300; sub index 1
	    pCan->sendData[3] = (unsigned char)subIndex;
	    if ( wFlag ) memcpy(&pCan->sendData[4],data,length);
//      if ( wFlag && length == 4 ) printf("can save data: %x\n",*(int *)&FastMsg.data[6]);
	    length = 8;
	    rtr = 0;
      break;
  	case PDO_TRANSMIT:
  	case PDO2_TRANSMIT:
	    memcpy(&pCan->sendData[0],data,length);
	    rtr=0;
	    break;
	  case PDO_DI_READ:
	  case PDO_AI_READ:
	    length = 0;
	    rtr = 1;
	    break;
	  default:
	    cdilog("canSendAPI: ****Wrong CanOpenID: 0x%x; line: %d\n",canOpenId, line+1);
	    return -3;
  }
  pCan->sendId = canOpenId + nodeId;				// can message used in can send
  pCan->sendRTR = rtr;
  pCan->sendLength = (char)length;
  pCan->xmtLength = 1;

  if ( NGdebug ) cdilog("CAN send id: 0x%x; len: 0x%x; rtr: %d; data: 0x%x 0x%x 0x%x 0x%x \n",pCan->sendId, pCan->sendLength, pCan->sendRTR, pCan->sendData[0], pCan->sendData[1], pCan->sendData[2], pCan->sendData[3]);

//  pCan->xmtStatus = canWrite(pCan->fd, &pCan->xmtMsg, &pCan->xmtLength, NULL);
//  return outCanTelegram(pCan);
  ret = sendCanTelegram(line, pCan->fd, pCan->sendId, pCan->sendRTR, pCan->sendLength, pCan->sendData, &(pCan->xmtStatus));
  if ( ret == -1 )
    cdilog("ESD CAN send Error! return code: 0x%x\n",pCan->xmtStatus); 
  else if ( ret == -2 )
    cdilog("PEAK CAN send Error! return code: 0x%x\n",pCan->xmtStatus); 
  if ( ret )
  {
    pCan->numSendErrors++;
    return -hardware_error;
  }
  return 0;
}

/****************************************/
/*	init can line						*/
/****************************************/
/**************************************/
void canMessage(int cdiLine, char *data, int leng, int retId, int rtr, int remoteNode, int remoteCommand)
{ 
  /* CdiLineInfoBlk *pLine = busHandle[cdiLine]; */
  CdiLineInfoBlk *busHandle;
  CdiRequestInfoBlk *pReq;
  short readFlag, returnObj;
  char *str;
  unsigned int datalen;
  char localData[8];

  /* find out device */
  if ((busHandle=cdiGetBusHandle(cdiLine)) == NULL)
  {
    cdilog("canMessage: Wrong cdi_line: %d for canMessage() call\n",cdiLine);
    return;
  }
  
  if (NGdebug) 
  {
    cdilog("CanMessage call Back: cdi line: %d, can line number: %d, data: %s, length: %d, retId: %d, rtr: %d\n",
            cdiLine, busHandle->busLine, data, leng, retId, rtr);
    cdilog("Number of active request: %d\n",busHandle->numberActiveRequest);
  }

  for ( busHandle->readCounter=0; busHandle->readCounter < busHandle->numberActiveRequest; busHandle->readCounter++ )
  {
    if ( NGdebug ) cdilog("enter loop ! Number of active request: %d; active flag = %d; preq addr: 0x%x\n",busHandle->numberActiveRequest, busHandle->pReqList[ busHandle->readCounter ].active, &busHandle->pReqList[ busHandle->readCounter ]);

    if ( (busHandle->pReqList[ busHandle->readCounter ].active != recvRequest &&
      busHandle->pReqList[ busHandle->readCounter ].active != recvRequest2nd) ||
      busHandle->pReqList[ busHandle->readCounter ].pDev->cdiAddr != remoteNode ) continue;


    pReq = &busHandle->pReqList[ busHandle->readCounter ];
    if ( pReq->active == recvRequest ) /* first part */
      readFlag = (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) ? 1 : 0;
    else /* second part! */
      readFlag = (pReq->accessFlag == cdiWriteReadFlag) ? 1 : 0;
    returnObj = *(short *)&data[1];

    datalen = leng;
    str = data;
    if ( pReq->devParameters[canOpenId] == SDO_TRANSMIT ) /* check messagesType */
    {
      if ( remoteCommand != SDO_RECEIVE ) continue;
      if ( pReq->devParameters[canOpenObj] != returnObj || (char)pReq->devParameters[canOpenSubIndex] != data[3] ) continue;
      if ( (data[0] & 0xf0) == 0x80 ) // sdo abort Domain Transfer
      {
        pReq->deviceError = *(int *)&data[4];
        if ( pReq->deviceError == 0x05040001 ) pReq->errorCode = illegal_mode;
        else if ( pReq->deviceError == 0x06010002 ) pReq->errorCode = illegal_operator;
        else if ( pReq->deviceError == 0x06020000 ) pReq->errorCode = invalid_index;
        else if ( pReq->deviceError == 0x06070010 ) pReq->errorCode = wrong_type;
        else if ( pReq->deviceError == 0x06090011 ) pReq->errorCode = illegal_equipment_number;
        else pReq->errorCode = not_defined;
        if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
        if (NGdebug) cdilog("SDO abort command: %x",pReq->deviceError);
        cdiExecRequest(&(busHandle->pReqList[busHandle->readCounter]), readTaskIndex);
        continue;
      }
      else if ( readFlag ) /* SDO read */
      {
        if ( (data[0] & 0xf0) != 0x40 ) continue;
        if ( (data[0] & 0xf) == 0xf ) datalen = 1;
        if ( (data[0] & 0xf) == 0xb ) datalen = 2;
        if ( (data[0] & 0xf) == 0x7 ) datalen = 3;
        if ( (data[0] & 0xf) == 0x3 ) datalen = 4;
        else datalen = 4;
      }
      else   /* SDO write */
      {
        if ( (data[0] & 0xf0) != 0x60 ) continue;
        datalen = leng - 4;
      }
      /* SDO data back.... */
      datalen = leng - 4;
      str = &data[4];
    }
    else if ( pReq->devParameters[canOpenReadId] != notdefinedMsg && pReq->devParameters[canOpenReadId] != remoteCommand ) continue; /* same OpenId? */
    else if ( readFlag == 0 ) /* for write command device should not answer! */
    {
      if (NGdebug) 
      {
        cdilog("canMessage Wrong: unexpected return message for write conmmand!");
        cdilog("CanOpenId: 0x%x; canLine: %d; module CanId: %d; module Name: %s",
                remoteCommand,busHandle->busLine,remoteNode,pReq->pDev->devName);
      }
      continue;
    }
    memcpy(localData, str, datalen);
    if ( readFlag ) /* check return condition */
    {
      if (NGdebug) cdilog("CAN data check condition for readFlag canOpenID: 0x%x",remoteCommand);
      checkReturnCondition(pReq, pReq->pInData, localData, datalen, 0);
      if ( pReq->dataReturnCondition == 0 ) continue; 
    }

    if (NGdebug) cdilog("can data return condition OK canOpenID: 0x%x",remoteCommand);
    
    memcpy( pReq->pInData, localData, datalen > pReq->inLength ? pReq->inLength : datalen);

    cdiWaitSemaphore(&busHandle->pReqList[busHandle->readCounter].exec_SemId,0);
    cdiExecRequest(&(busHandle->pReqList[busHandle->readCounter]), readTaskIndex);
    cdiPostSemaphore(&busHandle->pReqList[busHandle->readCounter].exec_SemId);
  }
}

#define CDI_PARAMETER_DELIMETER ":"
int cdiParseParamString(char *paramstr,int *parameters)
{
  char *s,*c=NULL;
  int i;
  if (paramstr == NULL || parameters == NULL) return argument_list_error;
  for (s=strtok_r(paramstr,CDI_PARAMETER_DELIMETER,&c), i=0; 
       i<NUMBER_EXEC_PARAM && s != NULL; 
       s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c), i++)
  {
    parameters[i] = atoi(s);
  }
  return 0;
}

CDI_CAN_EXPORT int initCanLineForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  int parameter[NUMBER_EXEC_PARAM];
  memset(parameter,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdiParseParamString(parameterList,parameter);
  return initCanLine(busLine,parameter,canMessage,cdiLine,numberDevices);
}

int initCanLine(int line, int *parameters, void (*callBack)(int, char *, int, int, int, int, int), int cdiLine, int numberDevices)
{ int ret, errorCode;
  CAN_LINE *pCan;
  DWORD dwCdiThreadId;
  struct timespec wait_time;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES )
  {
    cdilog("Open wrong Can line: %d; valid lines: 1 to %d\n",line, MAX_NUMBER_CAN_LINES);
    return -illegal_line;
  }
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd != 0 )
//#ifdef CDI_CAN_LINUX
//  if ( pCan->fd != 0 )
//#else
//  if ( pCan->fd != NULL )
//#endif
  {
    cdilog("Can line: %d already opened! \n",line+1);
    return 0;
  }

  pCan->canRate = transBaudRate( parameters[paramCanRate] );
  if ( pCan->canRate == -1 )
  {
    cdilog("Can line: %d; with Baudrate: %d not supported! \n",line,parameters[paramCanRate]);
    return -argument_list_error;
  }

  sprintf( pCan->devPath, "/dev/pcan%d", parameters[paramNodeNumber] );
  pCan->NumberDevices = numberDevices;
  pCan->cdiLine = cdiLine;
  pCan->receiveCanMessage = callBack;
  memset(&pCan->numRevMsg[0],0, sizeof(unsigned int) * numTypeMsg * MAX_CAN_IDS);

  pCan->txqueuesize = 8192; /*NTCAN_MAX_TX_QUEUESIZE; */
  pCan->rxqueuesize = 4096  /* NTCAN_MAX_RX_QUEUESIZE */;
  pCan->txtimeout = TXT_TIMEOUT;
  pCan->rxtimeout = 0;
  pCan->mode = 0;
  pCan->net = line;
  ret = canDrvOpen(line, pCan->devPath, pCan->canRate, pCan->net, pCan->mode, pCan->txqueuesize, pCan->rxqueuesize, pCan->txtimeout, pCan->rxtimeout, &(pCan->fd), &errorCode);
  if ( ret == -1 )
    cdilog("canOpen() failed with error: %d; ESD can line: %d\n",errorCode, line + 1);
  else if ( ret == -2 )
    cdilog("canSetBaudrate() failed with error: %d; ESD can line: %d\n",errorCode, line + 1);
  else if ( ret == -3 )
      cdilog("canIdAdd() failed with error: %d; ESD can line: %d\n",errorCode, line + 1);
  else if ( ret == -10 )
      cdilog("LINUX_CAN_Open() failed with error: %d; Peak can line: %d\n",errorCode, line + 1);
  else if ( ret == -11)
    cdilog("CAN_Init() failed with error: %d; Peak can line: %d\n",errorCode, line + 1);

  if ( ret ) return -file_not_open;

#ifdef CDI_CAN_LINUX
  if (pthread_create(&pCan->tIdMaster,NULL,GetCanDataThread,(void *)pCan) != 0)
#else
  if ( (pCan->tIdMaster = CreateThread(NULL,0,GetCanDataThread,(void *)pCan,0,&dwCdiThreadId)) == NULL )
#endif
  {
	  cdilog ("CreateThread foe CAN line: %d error",line + 1);
    terminateCanLine(line + 1);
	  return -un_allocated;
  }

  /* Sleep(1000); *//* wait for driver init finished, ready for sending */
  wait_time.tv_sec = 1;
  wait_time.tv_nsec = 0;
  nanosleep( &wait_time, NULL );

  pCan->numSendErrors = 0;
  pCan->statusOK = 0;
  cdilog("canOpen() OK; can line: %d\n", line + 1);

  return (int)pCan->fd;
}

/****************************************/
/*	terminate can line					*/
/****************************************/
CDI_CAN_EXPORT int terminateCanLine(int line)
{ CAN_LINE *pCan;
  int ret = 0;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd == 0 ) return -file_not_open;
//#ifdef CDI_CAN_LINUX
//  if ( pCan->fd == 0 ) return -file_not_open;
//#else
//  if ( pCan->fd == NULL ) return -file_not_open;
//#endif

  if ( pCan->tIdMaster)
  {
    cdiCancelThread((int)pCan->tIdMaster);
    pCan->tIdMaster = 0;
  }
  ret = canDrvClose( &(canLine[line].fd) );
  if ( ret != 0 ) /*  NTCAN_SUCCESS */
  {
    cdilog("canClose() failed with error: %d; can line: %d\n",ret, line + 1);
    ret = -file_not_open;    
  }
  else cdilog("canClose() OK; can line: %d\n",line + 1);

  return ret;
}
/****************************************/
/*	load Can card information			*/
/****************************************/
CDI_CAN_EXPORT int configCanCard(void)
{ int i;
  CAN_LINE *pCan;

  for ( i=0; i<MAX_NUMBER_CAN_LINES; i++ )
  {
    pCan = &(canLine[i]);
#ifdef JANZCAN_H_
    pCan->canRate = (i%2) == 0 ? BTR_500KB : BTR_20KB;
    pCan->canCardType = JANZ_PC_CARD;
#endif
    pCan->dpm_x = i/2;		// card number
    pCan->dpm_y = i%2;		// channel number
    pCan->lineNumber = i;
  }
  return 0;
}

