
/********************************************************/
/*	Janz CAN or other Can interface card for CDI				*/
/********************************************************/
#include "cdiCan.h" 	// Can card definition

int RegisterCanBus(void);
int canMsgSend(int line, short nodeId, short canOpenId, short length, char *data, short ObjNum, short subIndex, short wFlag);
int initCanLine(int line, int *parameters, void (*callBack)(int, char *, int, int, int, int, int), int cdiLine, int numberDevices);

CDI_CAN_EXPORT int filterCanbus(int *parameter, int accessFlag);
CDI_CAN_EXPORT void execCanbus(CdiRequestInfoBlk *pReq);
CDI_CAN_EXPORT int terminateCanLine(int line);
CDI_CAN_EXPORT int configCanCard(void);
CDI_CAN_EXPORT int initCanLineForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList);
CDI_CAN_EXPORT int canScanFcn(char *target,char *result,int resultBufferSize);

/* can card specific routines */
int canDrvOpen(int line, char *devPath, int canRate, int net, int mode, int txSize, int rxSize, int txTimeout, int rxTimeout, int *fd, int *erroCode);
int canDrvClose(int *fd);
int transBaudRate(int baudrate);
int readCanTelegram(int line, int fd, short *id, char *rtr, char *length, char *data);
int sendCanTelegram(int line, int fd, short sendId, char sendRTR, char sendLength, char *sendData, int *xmtStatus);

#ifdef JANZCAN_H_

#include "defs.h"	// Janz104 dpm headers
#include "vmod.h"
#include "icanif.h"
#include "calif.h"
#include "nmtm.h"
#include "dbtm.h"
#include "mcal.h"
#include "can_lib.h"
#include "dpm.h"
#include "mitop.h"

#endif /* JANZCAN_H_ */

#define MAX_NUMBER_CAN_LINES	4
#define MAX_CAN_IDS	    128       // # number of CanOpen IDs enabled for receive 0x7f

/*------------------------------------
 * CANopen commandos(ID) and objects
 *------------------------------------ */
#define SDO_TRANSMIT 0x600 /* request on remote node (read or write) */
#define SDO_RECEIVE 0x580 /* device back command on sdo_transmit(read or write) */
#define PDO_TRANSMIT 0x200 /* send without answer */
#define PDO2_TRANSMIT 0x300 /* send without answer */
#define PDO_DI_READ 0x180 /* digital input (read) */
#define PDO_AI_READ 0x280 /* analog input (read) */
#define NMT_COMMANDS 0x0
#define NMT_ERROS 0x700 /* from remote device */
#define EMERGENCE_MESSAGE 0x80 /* from remote device */

#define SDO_DIGI_OUTPUT_16_WR (SDO_TRANSMIT + 0x1) /* digital 16 bits out */
#define SDO_DIGI_OUTPUT_16_RD (SDO_TRANSMIT + 0x2) /* digital 16 bits read */
#define SDO_DIGI_OUTPUT_8_WR (SDO_TRANSMIT + 0x3) /* digital 8 bits out */
#define SDO_DIGI_OUTPUT_8_RD (SDO_TRANSMIT + 0x4) /* digital 8 bits read */

#define SDO_WRITE 0x22 /* sub command for SDO_TRANSMIT */
#define SDO_READ 0x40

#define SDO_DIGI_OUT_BYTE_OBJ 0x6200 /* object */
#define SDO_DIGI_OUT_OBJ 0x6300
#define SDO_TRANS_PDO_OBJ 0x1800
#define SDO_RECV_PDO_OBJ 0x1400

#define SDO_DESY_3000_OBJ 0x3000
#define SDO_DESY_3001_OBJ 0x3001

#define NMT_START_REMOTE 0x1         /* NMT data byte */
#define NMT_STOP_REMOTE 0x2
#define NMT_ENTER_PREOPERATION 0x80
#define NMT_RESET_NODE 0x81
#define NMT_RESET_COMMUNICATION 0x82

/* open CAN parameter index */
enum openCanParamIndex
{
  paramCanRate,     /* e.g 500 in KHz */
  paramNodeNumber,  /* 32 for USB peak can  */
};
enum execCanParamIndex
{
  msgType,   /* canMsgTypeIndex:   0, 1, 2, 3, 4, 5(any message) */
  canOpenObj, /* can open object number */
  canOpenSubIndex,
  canBroadcast, /* canID forced to be 0, one telegram for all devices */
  canAutoReceive, /* device sends data regulary (for Read or ReadthenWrite) */
  notused,
  canOpenId, /* can open id */
  canOpenReadId, /* can open id for read */
};

enum canMsgTypeIndex
{
  sdoMsg,
  pdo1Msg,
  pdo2Msg,
  nmtMsg,
  emergencyMsg,
  notdefinedMsg,
  numTypeMsg
};

//--------------------------
//	Janz Card spesific
//--------------------------
#define NUM_CAN_TYPE		3	// cancard type:
#define JANZ_PC_CARD		1	// pccard-ican3-cal
#define JANZ_PC_CARD_RAW	2	// pccard-ican3-rawCan
#define PC104_CARD			3	// pc 104

#define FILTER_LOW	0
#define FILTER_HIGH	0x7ff

#define	HI_PRIORITY		1  // for CAN104 priority queues
#define	MID_PRIORITY	2
#define	LO_PRIORITY		3

typedef struct {			/* can line	*/
  int	cdiLine;	// cdi buffer index, start from 0,1,2

  int	fd;							// can file descriptor
//#ifdef JANZCAN_H_
//  int	fd;							// can file descriptor
//#else
//  HANDLE	fd;							// can file descriptor
//#endif
  short	lineNumber, NumberDevices;	// can line number(>=0) and # of can modules
  char	devPath[20];				// driver path /dev/dpm_xy
  int	canRate;					// baudrate
  int	canCardType;				// PC card(CAL, RAW) or PC104
  short	dpm_x, dpm_y;				// modat number and can channels (0 or 1)

  long txqueuesize, rxqueuesize, txtimeout, rxtimeout;
  unsigned long mode, net;  // logical number

  short	statusOK;			// 1 initilized (at least one device alive)
  int	numSendErrors;		// number of can send errors within this line

  cdi_thread_t	tIdMaster;		// receive task variables
/*
#ifdef ESDCAN_H_
  CMSG cmsg[N_MSGS], *rcvPmsg, xmtMsg;
#endif
#ifdef PEAKCAN_H_
  TPCANMsg cmsg[N_MSGS], *rcvPmsg, xmtMsg;
#endif
  */
  unsigned long rcvStatus, recLength, recCnt, xmtLength;
  int xmtStatus;

  void (*receiveCanMessage)(int cdiLine, char *data, int length, int retId, int rtr, int remoteNode, int remoteCommand);

  short	sendId;				// can message used in can send
  char	sendRTR, sendLength;
  char	sendData[8];

  short	retId;				// can message used in receive thread
  char	rtr, length;
  char	retData[8];
  short remoteNode, remoteCommand;

  unsigned long numRevMsg[numTypeMsg][MAX_CAN_IDS];
} CAN_LINE;
