#include "cdiCan.h" 	// Can card definition

#include <stdio.h>
#include <stdlib.h>

#ifndef CDI_CAN_LINUX
# include "windows.h"
# include <time.h>

#else

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <memory.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <setjmp.h>

#include <math.h>
#include <stropts.h>	/* ioctl	*/
#include <semaphore.h>

#endif

extern void cdilog(char *text,...);

#ifdef ESDCAN_H_
#include "ntcan.h"
int n_baudrate = 14;
int esdBaudrate[] = {1000, 667, 500, 333, 250, 166, 125, 100, 67, 50, 33, 20, 13, 10};
#endif /* ESDCAN_H_ */

#ifdef PEAKCAN_H_
#if defined(CDI_CAN_LINUX)
#include <libpcan.h>
#else
#include "pcan_usb.h"
int saveCanRate[8];
#endif
#endif  /* PEAKCAN_H_ */

#define WAIT_RECQUEUE_MS  1      // 10ms

/**********************************/
int readCanTelegram(int line, int fd, short *id, char *rtr, char *length, char *data)
{ int errCode = 0, rcvStatus;
  struct tm *newtime;
  time_t aclock;

#if defined(ESDCAN_H_)
  CMSG cmsg;
  int recLength;
  do
  {
    recLength = 1;
#if defined(CDI_CAN_LINUX)
//    rcvStatus = canTake((NTCAN_HANDLE)fd, &cmsg, &recLength);
    rcvStatus = canRead((NTCAN_HANDLE)fd, &cmsg, &recLength, NULL);
#else
    rcvStatus = canRead((HANDLE)fd, &cmsg, &recLength, NULL);
#endif
    if ( rcvStatus == NTCAN_SUCCESS )
    {
//	  if ( recLength != 1 && rcvStatus != NTCAN_RX_TIMEOUT )
//	  {
//	    cdilog("ESD CAN canRead return more than 1 Telegram: %d! 0x%x, line: %d\n",recLength, rcvStatus, line+1);
//		printf("ESD CAN canRead return more than 1 Telegram: %d! 0x%x, line: %d\n",recLength, rcvStatus, line+1);
//	  }
	  if ( recLength == 1 )
	  {
	    *id = (short)cmsg.id;
		*rtr = (cmsg.len >> 4) & 0x1;
		*length = cmsg.len & 0xf;
		memcpy(data, cmsg.data, *length);
//		printf("id = %d; rtr = %d; length = %d\n", *id, *rtr, *length);
		return rcvStatus;
	  }
	  else
		rcvStatus = NTCAN_RX_TIMEOUT;
    }
    if ( rcvStatus != errCode && rcvStatus != NTCAN_RX_TIMEOUT) /* print once */
    {
      cdilog("ESD CAN receive Can error! 0x%x, line: %d\n",rcvStatus, line+1);
      printf("ESD CAN receive Can error! 0x%x, line: %d\n",rcvStatus, line+1);
      errCode = rcvStatus;
      if ( errCode == 21 )
      {
        printf("ESD CAN Controller (line number : %d) is disconnected!!! Please restart the server!!!\n",line+1);
        cdilog("ESD CAN Controller (line number : %d) is disconnected!!! Please restart the server!!!\n",line+1);
      }
    }
  } while ( rcvStatus != NTCAN_SUCCESS );
//  } while ( rcvStatus != NTCAN_SUCCESS && recLength != 1 );
#elif defined(PEAKCAN_H_)
  TPCANMsg cmsg;
  do
  {
#if defined(CDI_CAN_LINUX)
    if ((rcvStatus = CAN_Read((HANDLE)fd, &cmsg)))
    {
      perror("CDI PEAK CAN: CAN_Read()");
      cdilog("CDI PEAK CAN: CAN_Read() error: 0x%x, line: %d\n",rcvStatus, line+1);
    }
#else
    if ((rcvStatus = CAN_Read(&cmsg)) != CAN_ERR_OK)
    {
      if ( (rcvStatus & CAN_ERR_QRCVEMPTY) == CAN_ERR_QRCVEMPTY ) Sleep(WAIT_RECQUEUE_MS);
      else
      {
        printf("PEAK CAN_Read error! %d, cmsg.MSGTYPE: 0x%x; line: %d\n",rcvStatus, cmsg.MSGTYPE, line+1);
        cdilog("PEAK CAN_Read error! %d, cmsg.MSGTYPE: 0x%x; line: %d\n",rcvStatus, cmsg.MSGTYPE, line+1);
        errCode = rcvStatus;

        if ( cmsg.MSGTYPE & MSGTYPE_STATUS )
        {
          if ( cmsg.DATA[3] & CAN_ERR_OVERRUN )
          {
            printf("PEAK CAN_Read Status! CAN_ERR_OVERRUN! CAN Controller was read to late! line: %d\n", line+1);
            cdilog("PEAK CAN_Read Status! CAN_ERR_OVERRUN! CAN Controller was read to late! line: %d\n", line+1);
          }
          if ( cmsg.DATA[3] & CAN_ERR_BUSLIGHT )
          {
            printf("PEAK CAN_Read Status! CAN_ERR_BUSLIGHT! An error counter limit reached (96)! line: %d\n", line+1);
            cdilog("PEAK CAN_Read Status! CAN_ERR_BUSLIGHT! An error counter limit reached (96)! line: %d\n", line+1);
          }
          if ( cmsg.DATA[3] & CAN_ERR_BUSHEAVY )
          {
            printf("PEAK CAN_Read Status! CAN_ERR_BUSHEAVY! An error counter limit reached (128)! line: %d\n", line+1);
            cdilog("PEAK CAN_Read Status! CAN_ERR_BUSHEAVY! An error counter limit reached (128)! line: %d\n", line+1);
          }
          if ( cmsg.DATA[3] & CAN_ERR_BUSOFF )
          {
			time( &aclock );                 /* Get time in seconds */
		    newtime = localtime( &aclock );  /* Convert time to struct */
            printf("%s: PEAK CAN_Read Status! CAN_ERR_BUSOFF! Can Controller went Bus-Off! line: %d\n", asctime(newtime), line+1);
            cdilog("PEAK CAN_Read Status! CAN_ERR_BUSOFF! Can Controller went Bus-Off! line: %d\n", line+1);
            if ( (rcvStatus = CAN_Init((WORD)saveCanRate[line], CAN_INIT_TYPE_ST)) != CAN_ERR_OK )
            {
              printf("%s: PEAK CAN_Init error! error: 0x%x, line: %d\n", asctime(newtime), rcvStatus, line+1);
              cdilog("PEAK CAN_Init error! error: 0x%x, line: %d\n", rcvStatus, line+1);
            }
            rcvStatus = 1;
          }
        }
      }
    }
#endif
    else
    {
      *id = (short)cmsg.ID;
      *rtr = (cmsg.MSGTYPE & MSGTYPE_RTR) ? 1 : 0;
      *length = cmsg.LEN;
      memcpy(data, cmsg.DATA, *length);
    }
  } while ( rcvStatus );
#endif
  return 0;
}

/**********************************/
int sendCanTelegram(int line, int fd, short sendId, char sendRTR, char sendLength, char *sendData, int *xmtStatus)
{ int xmtLength = 1;

#if defined(ESDCAN_H_)
  CMSG cmsg;
  cmsg.id = sendId;
  cmsg.len = (sendRTR ? 0x10 : 0) + sendLength;
  memcpy(cmsg.data, sendData, 8);
#if defined(CDI_CAN_LINUX)
  *xmtStatus = canSend((NTCAN_HANDLE)fd, &cmsg, &xmtLength);
#else
  *xmtStatus = canSend((HANDLE)fd, &cmsg, &xmtLength);
#endif
  if ( *xmtStatus != NTCAN_SUCCESS ) return -1;
#elif defined(PEAKCAN_H_)
  TPCANMsg cmsg;
  cmsg.ID = sendId;
  cmsg.LEN = sendLength;
  cmsg.MSGTYPE = (sendRTR ? MSGTYPE_RTR : 0);
  memcpy(cmsg.DATA, sendData, 8);
#if defined(CDI_CAN_LINUX)
  *xmtStatus = CAN_Write((HANDLE)fd, &cmsg);
#else
  *xmtStatus = CAN_Write(&cmsg);
#endif
  if ( *xmtStatus ) return -2;
#endif
  return 0;
}

/**********************************/
int canDrvClose(int *fd)
{ int ret;

#if defined(ESDCAN_H_)
#if defined(CDI_CAN_LINUX)
  ret = canClose( *(NTCAN_HANDLE *)fd );
#else
  ret = canClose( *(HANDLE *)fd );
#endif
#elif defined(PEAKCAN_H_)
#if defined(CDI_CAN_LINUX)
  ret = CAN_Close( *(HANDLE *)fd );
#else
  ret = CAN_Close();
 	if ( ret ) 
  {
    printf(" Peak CAN_Close error: 0x%x\n",ret);
    cdilog(" Peak CAN_Close error: 0x%x\n",ret);
  }
#endif
#endif
  *fd = 0;
  return ret;
}

/**********************************/
int canDrvOpen(int line, char *devPath, int canRate, int net, int mode, int txSize, int rxSize, int txTimeout, int rxTimeout, int *fd, int *erroCode)
{ int ana;

#if defined(ESDCAN_H_)
  int id;
  txSize = NTCAN_MAX_TX_QUEUESIZE;
  rxSize = NTCAN_MAX_RX_QUEUESIZE;
//  rxTimeout = 1;
//  txSize = 10;
//  rxSize = 100;
//  txTimeout = 100;
//  rxTimeout = 10000;
#if defined(CDI_CAN_LINUX)
  ana = canOpen(net, mode, txSize, rxSize, txTimeout, rxTimeout, (NTCAN_HANDLE *)fd);
#else
  ana = canOpen(net, mode, txSize, rxSize, txTimeout, rxTimeout, (HANDLE *)fd);
#endif
  if ( ana != NTCAN_SUCCESS )
  {
    *erroCode = ana;
    return -1;  /*  canOpen() failed  */
  }
#if defined(CDI_CAN_LINUX)
  ana = canSetBaudrate(*(NTCAN_HANDLE *)fd, canRate);
#else
  ana = canSetBaudrate(*(HANDLE *)fd, canRate);
#endif
  if ( ana != 0 )
  {
    canDrvClose(fd);
    *erroCode = ana;
    return -2;  /*  SetBaudrate() failed  */
  }
  for (id=0; id<MAX_CAN_IDS; id++)
  {
#if defined(CDI_CAN_LINUX)
    ana = canIdAdd(*(NTCAN_HANDLE *)fd, id+SDO_RECEIVE);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(*(NTCAN_HANDLE *)fd, id+PDO_DI_READ);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(*(NTCAN_HANDLE *)fd, id+PDO_AI_READ);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(*(NTCAN_HANDLE *)fd, id+NMT_ERROS);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(*(NTCAN_HANDLE *)fd, id+EMERGENCE_MESSAGE);
#else
    ana = canIdAdd(*(HANDLE *)fd, id+SDO_RECEIVE);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(*(HANDLE *)fd, id+PDO_DI_READ);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(*(HANDLE *)fd, id+PDO_AI_READ);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(*(HANDLE *)fd, id+NMT_ERROS);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(*(HANDLE *)fd, id+EMERGENCE_MESSAGE);
#endif
    if ( ana != NTCAN_SUCCESS )
    {
      canDrvClose(fd);
      *erroCode = ana;
      return -3;  /*  canIdAdd() failed  */
    }
  }
#if defined(CDI_CAN_LINUX)
  int tout = 0;
  ana = canIoctl(*(NTCAN_HANDLE *)fd, NTCAN_IOCTL_GET_RX_TIMEOUT,&tout);
  if ( ana == NTCAN_SUCCESS )
  {
    printf(" ESDCAN ! devPath: %s; fd = 0x%x; canRate: 0x%x; receive timeout = %d(%d)\n",devPath,*fd, canRate, tout, rxTimeout);
    cdilog(" ESDCAN ! devPath: %s; fd = 0x%x; canRate: 0x%x; receive timeout = %d(%d)\n",devPath,*fd, canRate, tout, rxTimeout);
  }
#endif
#elif defined(PEAKCAN_H_)
#if defined(CDI_CAN_LINUX)
  char txt[VERSIONSTRING_LEN];
	printf(" Peak CAN_Open and init! devPath: %s; fd = 0x%x; canRate: 0x%x\n",devPath,*fd, canRate);
	cdilog(" Peak CAN_Open and init! devPath: %s; fd = 0x%x; canRate: 0x%x\n",devPath,*fd, canRate);
  *fd = (int)LINUX_CAN_Open(devPath, O_RDWR);
  if ( *fd == 0 )
  {
    *erroCode = nGetLastError();
    return -10;    
  }
  errno = CAN_VersionInfo((HANDLE)(*fd), txt);
  if (!errno)
    printf("Peak Can: driver version = %s\n", txt);
  else {
    perror("Error!! Peak Can: CAN_VersionInfo()");
    cdilog("Error!! Peak Can: CAN_VersionInfo(): 0x%x",errno);
  }
  errno = CAN_Init((HANDLE)(*fd), canRate, CAN_INIT_TYPE_ST);
  if (errno) {
    perror("CDI PEAK CAN: CAN_Init()");
 		printf(" CAN_Init error: 0x%x, devPath: %s; fd = 0x%x; canRate: 0x%x\n",errno,devPath,*fd, canRate);
 		cdilog(" CAN_Init error: 0x%x, devPath: %s; fd = 0x%x; canRate: 0x%x\n",errno,devPath,*fd, canRate);
    CAN_Close( (HANDLE)(*fd) );
    *fd = 0;
    *erroCode = errno;
    return -11;    
  }
#else
	printf(" Peak CAN_Init! canRate: 0x%x\n",canRate);
  if ( (ana = CAN_Init((WORD)canRate, CAN_INIT_TYPE_ST)) != CAN_ERR_OK )
  {
 		printf(" Peak CAN_Init error: 0x%x; canRate: 0x%x\n",ana, canRate);
 		cdilog(" Peak CAN_Init error: 0x%x; canRate: 0x%x\n",ana, canRate);
    *erroCode = ana;
    return -1;  /*  canOpen() failed  */
  }
  saveCanRate[line] = canRate;
  *fd = 1;
#endif
#endif

  return 0;
}

/**********************************/
int transBaudRate(int baudrate)
{ int rate;

#if defined(ESDCAN_H_)
  for ( rate=0; rate<n_baudrate; rate++ ) if ( baudrate == esdBaudrate[rate] ) break;
  return rate < n_baudrate ? rate : -1;

#elif defined(PEAKCAN_H_)
  switch ( baudrate )		// set baud rate
  {
  	case 1000:	rate = CAN_BAUD_1M;		break;
  	case 500:	rate = CAN_BAUD_500K;		break;
  	case 250:	rate = CAN_BAUD_250K;		break;
    case 125:	rate = CAN_BAUD_125K;		break;
  	case 100:	rate = CAN_BAUD_100K;		break;
  	case 50:	rate = CAN_BAUD_50K;		break;
  	case 20:	rate = CAN_BAUD_20K;		break;
  	case 10:	rate = CAN_BAUD_10K;		break;
  	case 5:	rate = CAN_BAUD_5K;		break;
  	default:
	  	printf(" peak can baud rate not supported: %d",baudrate);
  		return -1;
  }
  return rate;

#else
  switch ( baudrate )		// set baud rate
  {
  	case 800:		rate = BTR_800KB;		break;
  	case 500:		rate = BTR_500KB;		break;
  	case 250:		rate = BTR_250KB; 	break;
  	case 125:		rate = BTR_125KB;		break;
  	case 50:		rate = BTR_50KB;		break;
  	case 20:		rate = BTR_20KB;		break;
  	case 10:		rate = BTR_10KB;		break;
  	case 1000:	rate = BTR_1MB;		break;
  	default:
  		printf(" baud rate not supported: %d",baudrate);
	  	return -1;
   }
   return rate;
#endif
}
