#include <stdio.h>
#include <stdlib.h>

#include "tinecdi.h"
#include "dbtools.h"
#include "cdi.h"	// interface to user program
#include "cdiCanLib.h"

#ifndef CDI_CAN_LINUX
# include "windows.h"
#define snprintf _snprintf

#else

//#include <stdio.h>
//#include <stdlib.h>
#include <errno.h>
#include <unistd.h>   // exit
#include <signal.h>
#include <string.h>
#include <stdlib.h>   // strtoul
#include <fcntl.h>    // O_RDWR
#include <ctype.h>
/*
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <memory.h>
#include <fcntl.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <setjmp.h>

#include <math.h>
#include <stropts.h>
#include <semaphore.h>
*/
#endif

char *canMsgTypeString[] = {"SDO: ","PDO1: ","PDO2: ","PDO3: ","PDO4: ","NMT: ","EMERGENCY: ","NotDefined: "};

char *scanCommand[] = {"getcommands","getlines","getrcvdmsg","getmodules","readsdo","writesdo",
      "readpdo1","writepdo1","readpdo2","writepdo2","readpdo3","writepdo3","readpdo4","writepdo4","writenmt", "readnmt",
      "startremote","stopremote","enterpreoperation","resetnode","resetcommunication"};
enum scommand {getcommands, getlines, getrcvdmsg, getmodules, readsdo, writesdo,
      readpdo1, writepdo1, readpdo2, writepdo2, readpdo3, writepdo3, readpdo4, writepdo4, writenmt, readnmt,
      startremote, stopremote, enterpreoperation, resetnode, resetcommunication, nCommands};

#define MAX_ASYNCOUNTS	140	// each asyn ignor delay 7 ms, total 1 seconds
#define DELAY_ASYNC_MS	7

#define WAIT_CANMSG_MS  600000000      // 600ms
#define WAIT_SDO_MS     10000000      // 10ms
#define WAIT_SDO_TOTOL_MS     800//300//150//100   // 500ms 2009-1 change to 600 ms; 2013-7 change to 700; 2015-11 change to 100ms
#define WAIT_SDO_TOTOL_MS_ALL 5500//6500//2500//1000      // 900ms; 2013-7 change to 1000
#define WAIT_SDO_SINGLE_MS    700//200//400       // maximum wait time for SDO message back 2016-1

#define WAIT_SDO_NOBLOCK     7//7      // ms time between teo SDO send telegrams for nonBlock transfere

#define NUMBER_WAIT_SDO       20      // number of WAIT_SDO_MS for SDO waiting
#define TXT_TIMEOUT     200           // ms
char *scanBuffer;
unsigned int scanBufferSize;
int scanFlag = 0;
int scanNodId = 0;

unsigned int numStoredMsg[MAX_CAN_IDS];
sem_t scan_SemId;

#ifdef CDI_CAN_LINUX
__attribute__((constructor)) void cdiCanMain(void)
{
  int cc = 0;

  if ((cc=cdiRegisterBusFilter(CAN_LIB_NAME,filterCanbus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusConfiguration(CAN_LIB_NAME,configCanCard)) != 0) goto err;
  if ((cc=cdiRegisterBusInitialization(CAN_LIB_NAME,initCanLineForCDI)) != 0 ) goto err;
  if ((cc=cdiRegisterBusHandler(CAN_LIB_NAME,execCanbus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusCleanup(CAN_LIB_NAME,terminateCanLine)) != 0 ) goto err;
  if ((cc=cdiRegisterBusScanner(CAN_LIB_NAME,canScanFcn)) != 0 ) goto err;
  if ((cc=cdiRegisterBusReqTimeout(CAN_LIB_NAME,canTimeoutFcn)) != 0 ) goto err;
  if ((cc=cdiRegisterBusDeviceReady(CAN_LIB_NAME,canDeviceReadyFcn)) != 0 ) goto err;
  if ((cc=cdiRegisterBusLoadRequest(CAN_LIB_NAME,canLoadRequest)) != 0 ) goto err;

err:
  if (cc)
  {
    cdilog("%s : failure in bus registration %s",erlst[cc],CAN_LIB_NAME);
  }
  return;
}
#endif

int NGdebug = 0;
int canMsgPrint = 0;
#define MAX_BLEN 256

CAN_LINE canLine[MAX_NUMBER_CAN_LINES] = {
	{-1,0}, {-1,0}, {-1,0}, {-1,0}
//#ifdef CDI_CAN_LINUX
//	{-1,0}, {-1,0}, {-1,0}, {-1,0}
//#else
//	{-1,NULL}, {-1,NULL}, {-1,NULL}, {-1,NULL}
//#endif
};

CDI_CAN_EXPORT int canScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256], canData[8];
  int i, j, cnt, k, max, ret, slen = 0, tmpId;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;
  struct timespec wait_time;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0 ) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0); //atoi(c);

  if (j == getcommands)
  {
    snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }  
  
  if ( j == getlines )
  {
    snprintf(result,resultBufferSize,"CAN BUS lines configured:\n");
    for (i=0, cnt=0; i<MAX_NUMBER_CAN_LINES; i++)
    {
      if ( canLine[i].cdiLine < 0 ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
       snprintf(&result[slen],resultBufferSize-slen,"%d",i+1);
      cnt++;
    }
    return 0;
  }
  
  line = data[0];
  if ( nData < 1 || line < 1 || line > MAX_NUMBER_CAN_LINES || canLine[line-1].cdiLine < 0) return no_such_line;
  scanBufferSize = resultBufferSize;
  scanBuffer = result;

  if ( j == getrcvdmsg )
  {
    crate = ( nData > 1 && data[1] > 0 && data[1] < MAX_CAN_IDS ) ? data[1] : 0;
    if ( crate ) { i = crate; max = crate + 1; }
    else { i = 0; max = MAX_CAN_IDS; };
    snprintf(result,resultBufferSize,"received CAN messages on line %d:\n",line);
    for (; i<max; i++)
    {
      for (k=0, cnt=0; k<numTypeMsg; k++)
      {
        if ( canLine[line-1].numRevMsg[k][i] == 0 ) continue;
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
        if ( cnt > 0 )
        {
          if ((slen = strlen(result)) > resultBufferSize - 1) break;
          snprintf(&result[slen],resultBufferSize-slen,", ");
        }
        else
        {
          if ((slen = strlen(result)) > resultBufferSize - 15) break;
          snprintf(&result[slen],resultBufferSize-slen,"NodId:%d received:\n",i);
        }
        if ((slen = strlen(result)) > resultBufferSize - 10) break;
        snprintf(&result[slen],resultBufferSize-slen,"%s",canMsgTypeString[k]);
        if ((slen = strlen(result)) > resultBufferSize - 4) break;
        snprintf(&result[slen],resultBufferSize-slen,"%d",canLine[line-1].numRevMsg[k][i]);
        cnt++;
      }
      if ( cnt > 0 )
      {
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
        snprintf(&result[slen],resultBufferSize-slen,"\n");
      }
    }
    return 0;
  }
  if ( j >= startremote && j <= resetcommunication )
  {
    crate = ( nData > 1 && data[1] > 0 && data[1] < MAX_CAN_IDS ) ? data[1] : 0;
    snprintf(result,resultBufferSize,"%s on line %d for Node: %d\n",scanCommand[j], line, crate);

    scanFlag = j;
    scanNodId = 0;

    if ( j == startremote ) wData = NMT_START_REMOTE;
    else if ( j == stopremote ) wData = NMT_STOP_REMOTE;
    else if ( j == enterpreoperation ) wData = NMT_ENTER_PREOPERATION;
    else if ( j == resetnode ) wData = NMT_RESET_NODE;
    else wData = NMT_RESET_COMMUNICATION;

  	if ( ret = canMsgSend(line, crate, NMT_COMMANDS, 1, (char *)&wData, 0, 0, 0, 0)  )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
    }
    else
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = WAIT_CANMSG_MS; 
      cdi_nanosleep( &wait_time, NULL );
    }
    scanFlag = 0;
    return 0;
  }
  if ( j == getmodules )  // request pdos
  {
    snprintf(result,resultBufferSize,"NodIDs connected on line %d:\n",line);
    for (i=0; i<MAX_CAN_IDS; i++) numStoredMsg[i] = canLine[line-1].numRevMsg[sdoMsg][i];
    // request sdo index 0x1000
  	for (  crate=1; crate<MAX_CAN_IDS; crate++ )
    {
	  if ( ret = canMsgSend(line, crate, SDO_TRANSMIT, 0, NULL, (short)0x1000, (short)0, 0, 0) )
      {
        if ((slen = strlen(result)) < resultBufferSize - 20)
          snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
        return 0;
      }
    }
    wait_time.tv_sec = 0;
    wait_time.tv_nsec = WAIT_CANMSG_MS; 
	cdi_nanosleep( &wait_time, NULL );
    for (i=0, cnt=0; i<MAX_CAN_IDS; i++)
    {
      if ( canLine[line-1].numRevMsg[sdoMsg][i] <= numStoredMsg[i] ) continue;
      if ( cnt > 0 )
      {
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
        snprintf(&result[slen],resultBufferSize-slen,",");
      }
      if ((slen = strlen(result)) > resultBufferSize - 4) break;
      snprintf(&result[slen],resultBufferSize-slen,"%d",i);
      cnt++;
    }
	return 0;

#ifdef RTR_PDO_2008_11

    for (i=0; i<MAX_CAN_IDS; i++) numStoredMsg[i] = canLine[line-1].numRevMsg[pdo1Msg][i] + canLine[line-1].numRevMsg[pdo2Msg][i] + canLine[line-1].numRevMsg[pdo3Msg][i] + canLine[line-1].numRevMsg[pdo4Msg][i];

//    shared = value = 0;
//    if ((retSem=cdiGetSemaphore(&scan_SemId,shared,value)) != 0)
//    {
//      if ((slen = strlen(result)) < resultBufferSize - 20)
//        snprintf(&result[slen],resultBufferSize-slen,"GetSemaphore Error!");
//      return 0;
//    }
  	
    // start remote to use pdos
    crate = 0;
    wData = NMT_START_REMOTE;

  	if ( ret = canMsgSend(line, crate, NMT_COMMANDS, 1, (char *)&wData, 0, 0, 0, 0) )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
      return 0;
    }
    // request pdo1 and pdo2
  	for (  crate=1; crate<MAX_CAN_IDS; crate++ )
    {
	  ret = canMsgSend(line, crate, PDO_DI_READ, 0, NULL, 0, 0, 0, 0);
      ret = canMsgSend(line, crate, PDO_AI_READ, 0, NULL, 0, 0, 0, 0);
      ret = canMsgSend(line, crate, PDO3_READ, 0, NULL, 0, 0, 0, 0);
      ret = canMsgSend(line, crate, PDO4_READ, 0, NULL, 0, 0, 0, 0);
      if ( ret = canMsgSend(line, crate, NMT_COMMANDS, 1, (char *)&wData, 0, 0, 0, 0) )
      {
        if ((slen = strlen(result)) < resultBufferSize - 20)
          snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
        return 0;
      }
    }

    wait_time.tv_sec = 0;
    wait_time.tv_nsec = WAIT_CANMSG_MS; 
    cdi_nanosleep( &wait_time, NULL );
//    cdiWaitSemaphore(&scan_SemId,1);
//printf("wait finished!\n");
    for (i=0, cnt=0; i<MAX_CAN_IDS; i++)
    {
      if ( canLine[line-1].numRevMsg[pdo1Msg][i] + canLine[line-1].numRevMsg[pdo2Msg][i] + canLine[line-1].numRevMsg[pdo3Msg][i] + canLine[line-1].numRevMsg[pdo4Msg][i] <= numStoredMsg[i] ) continue;
      if ( cnt > 0 )
      {
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
        snprintf(&result[slen],resultBufferSize-slen,",");
      }
      if ((slen = strlen(result)) > resultBufferSize - 4) break;
      snprintf(&result[slen],resultBufferSize-slen,"%d",i);
      cnt++;
    }
//    cdiDestroySemaphore(&scan_SemId);
    return 0;
#endif
  }

  if ( nData < 2 || data[1] < 1 || data[1] >= MAX_CAN_IDS ) return argument_list_error;
  crate = data[1];
  if ( j == readpdo1 || j == readpdo2 || j == readpdo3 || j == readpdo4 || j == readnmt)  // read pdo & nmt messages
  {
    snprintf(result,resultBufferSize,"%s on line %d for Node: %d\n",scanCommand[j], line, crate);
    scanFlag = j;
    scanNodId = crate;
	if ( j != readnmt )
	{
	  if ( j == readpdo1 ) tmpId = PDO_DI_READ;
	  else if ( j == readpdo2 ) tmpId = PDO_AI_READ;
	  else if ( j == readpdo3 ) tmpId = PDO3_READ;
	  else tmpId = PDO4_READ;
	}
    if ( j != readnmt && (ret = canMsgSend(line, crate, (short)tmpId, 0, NULL, 0, 0, 0, 0)) )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
    }
    else
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = WAIT_CANMSG_MS; 
      cdi_nanosleep( &wait_time, NULL );
    }
    scanFlag = 0;
    scanNodId = 0;
    return 0;
  }

  if ( nData < 3 ) return argument_list_error;
  if ( j == writepdo1 || j == writepdo2 || j == writepdo3 || j == writepdo4 || j == writenmt)  // write pdo and nmt messages
  {
    snprintf(result,resultBufferSize,"%s on line %d for Node: %d\n",scanCommand[j], line, crate);
    for (i=2; i<nData && i<10; i++ ) canData[i-2] = (char)data[i];
    rData = i - 2;

    if ( j == writepdo1 ) subaddr = PDO_TRANSMIT;  
    else if ( j == writepdo2 ) subaddr = PDO2_TRANSMIT;  
    else if ( j == writepdo3 ) subaddr = PDO3_TRANSMIT;  
    else if ( j == writepdo4 ) subaddr = PDO4_TRANSMIT;  
    else subaddr = NMT_COMMANDS;

    scanFlag = j;
    scanNodId = crate;
    if ( ret = canMsgSend(line, crate, subaddr, rData, canData, 0, 0, 1, 0) )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
    }
    else
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = WAIT_CANMSG_MS; 
      cdi_nanosleep( &wait_time, NULL );
    }    
    scanFlag = 0;
    scanNodId = 0;
    return 0;
  }

  if ( nData < 4 ) return argument_list_error;
  if ( j == readsdo || j == writesdo)  // read sdo
  {
    for (i=4; i<nData && i<8; i++ ) canData[i-4] = (char)data[i];
    rData = i - 4;
    if ( j == writesdo && rData < 1 ) return argument_list_error;

    scanFlag = j;
    scanNodId = crate;
    if ( ret = canMsgSend(line, crate, SDO_TRANSMIT, rData, canData, (short)data[2], (short)data[3], (short)(j == readsdo ? 0 : 1), 0) )
    {
      if ((slen = strlen(result)) < resultBufferSize - 20)
        snprintf(&result[slen],resultBufferSize-slen,"CAN SendError: line: %d, Id: %d, Error: 0x%x",line, crate, ret);
    }
    else
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = WAIT_CANMSG_MS; 
      cdi_nanosleep( &wait_time, NULL );
    }
    scanFlag = 0;
    scanNodId = 0;
    return 0;
  }

  return 0;
}

CDI_CAN_EXPORT int filterCanbus(int *parameter, int accessFlag)
{
  if (parameter == NULL) return argument_list_error;
  if (parameter[msgType] < sdoMsg || parameter[msgType] > notdefinedMsg ) return argument_list_error;

  /* only read possible */
  if (accessFlag != cdiReadFlag && (parameter[msgType] == emergencyMsg || parameter[msgType] == notdefinedMsg) ) return argument_list_error;
  
  /* set canAutoReceive flag */
  if (parameter[msgType] == emergencyMsg || parameter[msgType] == notdefinedMsg) parameter[canAutoReceive] = 1;
  else if (parameter[msgType] == nmtMsg && accessFlag != cdiWriteFlag) parameter[canAutoReceive] = 1;

  /* set canOpenId & canOpenReadId */
  if ( parameter[msgType] == sdoMsg ) { parameter[canOpenId] = SDO_TRANSMIT;  parameter[canOpenReadId] = SDO_TRANSMIT;}
  else if ( parameter[msgType] == pdo1Msg ) { parameter[canOpenId] = PDO_TRANSMIT;  parameter[canOpenReadId] = PDO_DI_READ;}
  else if ( parameter[msgType] == pdo2Msg ) { parameter[canOpenId] = PDO2_TRANSMIT;  parameter[canOpenReadId] = PDO_AI_READ;}
  else if ( parameter[msgType] == pdo3Msg ) { parameter[canOpenId] = PDO3_TRANSMIT;  parameter[canOpenReadId] = PDO3_READ;}
  else if ( parameter[msgType] == pdo4Msg ) { parameter[canOpenId] = PDO4_TRANSMIT;  parameter[canOpenReadId] = PDO4_READ;}
  else if ( parameter[msgType] == nmtMsg ) { parameter[canOpenId] = NMT_COMMANDS;  parameter[canOpenReadId] = NMT_ERROS;}
  else if ( parameter[msgType] == emergencyMsg ) { parameter[canOpenId] = NMT_COMMANDS;  parameter[canOpenReadId] = EMERGENCE_MESSAGE;}
  else { parameter[canOpenId] = NMT_COMMANDS;  parameter[canOpenReadId] = notdefinedMsg;} /* any messages */

  return 0;
}

/****************************************************************************/
int sdoBlockTransfer(CdiRequestInfoBlk *pReq, char *data, int stateFlag)
{ char protoByte;
  int canRet = 0;

  if (stateFlag == 0)   /* start: init block transfer  */
  {
    protoByte = 0x20; /*  initiate SDO Download or Upload */
//  canRet = canSdoBlockMsgSend(protoByte);
    if ( canRet == 0 )
    {
      pReq->devParameters[downloadFlag] = *(int *)data;
      pReq->devParameters[dataIndex] = 0;
      pReq->devParameters[transState] = 1;
    }
    return canRet;
  }
  if ( pReq->devParameters[transState] == 1 ) /* confirmation */
  {
    if ( pReq->devParameters[downloadFlag] )  /* write  */
    {
      if ( data[0] == 0x60 ) pReq->devParameters[transState] = 2;
//      else canRet = abort_transfer;
    }
    else            /* read */
    {
    }
  }
  else if ( pReq->devParameters[transState] == 2 ) /* upload or downlod data */
  {
  }
  return 0;
}

/****************************************************************************/
int getSDOindex(CdiRequestInfoBlk *pReq)
{ int firstPart, returnIndex;

  if ( pReq->devParameters[canOpenId] != SDO_TRANSMIT ) return 0;
  if ( pReq->active == recvRequest ) firstPart = 1;         /* first transfere  */
  else if ( pReq->active == recvRequest2nd ) firstPart = 0; /* second transfere  */
  else
  {
    cdilog("canMsgSend parameters inconsistant: active state: %d; pReq = 0x%x! SDO: indexes: 0x%x, 0x%x; line: %d\n",pReq->active, pReq, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex], ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine);
    return 0;
  }
  returnIndex = 0;    /* read and write might have tifferent SDO indexes  */
  if ( firstPart )
  {
    if ( pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag ) returnIndex = 1;
  }
  else
  {
    if ( pReq->accessFlag == cdiReadWriteFlag ) returnIndex = 1;
  }
  if ( returnIndex ) if ( pReq->devParameters[canBroadcast] == 0 ) returnIndex = 0;
  return returnIndex;
}

/****************************************************************************/
int sendCdiCanMessage(CdiRequestInfoBlk *pReq, int wFlag)
{ int canRet=0, sdoCmd, index, nodeId, ln;
  CAN_LINE *pCan;

  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
  pCan = &canLine[ln];

  index = getSDOindex(pReq);
  sdoCmd = (wFlag && pReq->devParameters[canOpenId] == SDO_TRANSMIT && pReq->outLength > 4) ? SDO_WRITE_INIT_SEGMENT : 0;
  nodeId = (pReq->devParameters[canOpenId] == SDO_TRANSMIT || pReq->devParameters[canBroadcast] == 0) ? pReq->pDev->cdiAddr[caAddrIndex] : 0;
  if (pReq->devParameters[msgType] == sdoMsg && pCan->sdoNoBlockFlag )	//2009-9
  {
	if ( sdoCmd ) canRet = -illegal_data_size;
  }
  if ( canRet == 0 )
  canRet = canMsgSend(ln+1,
//              (short)(pReq->devParameters[canBroadcast] ? 0 : pReq->pDev->cdiAddr[caAddrIndex]),
              (short)nodeId,
              (short)pReq->devParameters[wFlag ? canOpenId : canOpenReadId], 
              (short)(sdoCmd ? 4 : pReq->outLength), 
              (char *)(sdoCmd ? &pReq->outLength : pReq->pOutData),
              (short)pReq->devParameters[index ? canBroadcast : canOpenObj], 
              (short)pReq->devParameters[index ? canAutoReceive : canOpenSubIndex], 
              (short)wFlag,
              (short)sdoCmd);

  if ( NGdebug )
    cdilog("canMsgSend return code: %d! line: %d\n",canRet, ln+1);
  
  pReq->devParameters[transState] = 0;
  if ( canRet < 0 )
  {
//    if (pReq->devParameters[msgType] == sdoMsg)
    if (pReq->devParameters[msgType] == sdoMsg && pCan->sdoNoBlockFlag == 0 )	//2009-9
    {
      cdiMutexLock(&pCan->sdoMutex[nodeId]);
      pCan->segmentFlag[nodeId] = 0;
      pCan->reqBufPointer[nodeId] = NULL;
      cdiMutexUnlock(&pCan->sdoMutex[nodeId]);
    }
    pReq->errorCode = -canRet; /* write failed */
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( pReq->errorCode == hardware_error )
    {
      pReq->deviceError = net_write_error;    // can write error
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest;
  }
  // 2009-9
  else if (pReq->devParameters[msgType] == sdoMsg && pCan->sdoNoBlockFlag )	//2009-9
  {
	pCan->sdoRequstTime[pReq->pDev->cdiAddr[caAddrIndex]] = getTimeMilliseconds();
  }
  return canRet;
}
/****************************************************************************/
CDI_CAN_EXPORT int canTimeoutFcn(CdiRequestInfoBlk *pReq, int set, int taskIndex)
{ int tmoutFlag;
  int canNod, ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
  CAN_LINE *pCan = &canLine[ln];

//  if ( pReq->devParameters[msgType] != sdoMsg ) return 0;
  if ( pReq->devParameters[msgType] != sdoMsg || pCan->sdoNoBlockFlag ) return 0;	// 2009-9
  canNod = pReq->pDev->cdiAddr[caAddrIndex];
  cdiMutexLock(&pCan->sdoMutex[canNod]);
  if ( pCan->segmentFlag[canNod] == 0 ||
      pCan->reqBufPointer[canNod] != pReq ||
      (pReq->active != recvRequest && pReq->active != recvRequest2nd) )
  {
    if ( set == 1 && pCan->segmentFlag[canNod] && (pReq->active != recvRequest && pReq->active != recvRequest2nd) )
      cdilog("SDO timeout function! line: %d; preq addr: 0x%x; NodeAddr: 0x%x; sdoRequstTime: %d; sdoReceiveTime: %d; SegmentFlag: %d; transState: %d; NodeId: %d; activeMode=%d; taskIndex: %d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, pCan->reqBufPointer[canNod], pCan->sdoRequstTime[canNod], pReq->devParameters[sdoMsgTiStIndex], pCan->segmentFlag[canNod], pReq->devParameters[transState], canNod, pReq->active, taskIndex, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);

    cdiMutexUnlock(&pCan->sdoMutex[canNod]);
    return 0;
  }
//  if ( abs(getTimeMilliseconds() - pCan->sdoRequstTime[canNod]) < WAIT_SDO_TOTOL_MS_ALL ) tmoutFlag = 0;
  if ( getTimeMilliseconds() - pCan->sdoRequstTime[canNod] < WAIT_SDO_SINGLE_MS ) tmoutFlag = 0;
  else tmoutFlag = 1;
  if ( set == 1 )
  {
    tmoutFlag = 1;
    cdilog("SDO timeout function! line: %d; preq addr: 0x%x; sdoRequstTime: %d; SegmentFlag: %d; NodeId: %d; activeMode=%d; taskIndex: %d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, pCan->sdoRequstTime[canNod], pCan->segmentFlag[canNod], canNod, pReq->active, taskIndex, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
//  if ( pReq->devParameters[msgType] == sdoMsg )
//    pCan->segmentFlag[canNod] = 0;
    pReq->devParameters[transState] = 0;
    pCan->segmentFlag[canNod] = 0;
    pCan->reqBufPointer[canNod] = NULL;

//    cdilog("SDO timeout function! line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; taskIndex: %d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pReq->active, taskIndex, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
  }
  cdiMutexUnlock(&pCan->sdoMutex[canNod]);
  return tmoutFlag;
}
/****************************************************************************/
CDI_CAN_EXPORT int canDeviceReadyFcn(CdiRequestInfoBlk *pReq, int taskIndex)
{ struct timespec wait_time;
  int diff, canNod, ln;
  CAN_LINE *pCan;

  if ( pReq->devParameters[msgType] != sdoMsg ) return 1;
  canNod = pReq->pDev->cdiAddr[caAddrIndex];
  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
  pCan = &canLine[ln];

  // background task: avoid sdoMsg request semaphore
  if ( pCan->segmentFlag[canNod] > 0 && taskIndex != execIndex && (getTimeMilliseconds() - pReq->devReadyTime) < WAIT_SDO_TOTOL_MS_ALL ) return 0;

  if ( cdiWaitSemaphore(&pCan->node_SemId, taskIndex != execIndex ? 1 : 0) )  // background task uses samephore trywait
  {
    if ( taskIndex == execIndex )
    {
      pReq->errorCode = semaphore_error;
      if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      pReq->active = finishRequest; /* finish request */
      cdilog("canDeviceReadyFcn : semaphore ERROR! line: %d, can node: %d, segmenFlag: %d, taskIndex:%d", ln+1, canNod, pCan->segmentFlag[canNod], taskIndex);
    }
    //cdilog("canDeviceReadyFcn : semaphore ERROR! line: %d, can node: %d segmenFlag:%d, taskIndex:%d", ln+1, canNod, pCan->segmentFlag[canNod], taskIndex);
    return 0;
  }

  if ( pCan->sdoNoBlockFlag )	// 2009-9
  {
	diff = abs(getTimeMilliseconds() - pCan->sdoRequstTime[canNod]);
	if ( diff < WAIT_SDO_NOBLOCK )
	{
		wait_time.tv_sec = 0;
	    wait_time.tv_nsec = (WAIT_SDO_NOBLOCK - diff) * 1000000;	// ns
		cdi_nanosleep( &wait_time, NULL );
	}
	return 1;
  }
  if ( pCan->segmentFlag[canNod] == 0 ) return 1;
  if ( pReq->devReadyTime == 0 ) pReq->devReadyTime = getTimeMilliseconds();
  // 2013-7 wait till, after the sdo timeout plus extra 600ms to avoid SDO segment flag on
//  if ( taskIndex != execIndex ) return (getTimeMilliseconds() - pReq->devReadyTime) > WAIT_SDO_TOTOL_MS ? 1 : 0;
//  if ( taskIndex != execIndex ) return (getTimeMilliseconds() - pReq->devReadyTime) > (WAIT_SDO_TOTOL_MS_ALL + 600) ? 1 : 0;

  if ( taskIndex != execIndex )
  {
//    return (getTimeMilliseconds() - pReq->devReadyTime) > (WAIT_SDO_TOTOL_MS_ALL + 600) ? 1 : 0;
    if ( (getTimeMilliseconds() - pReq->devReadyTime) > (WAIT_SDO_TOTOL_MS_ALL + 600 + pCan->asynIgnorCounter[canNod] * DELAY_ASYNC_MS) )
    {
	  if ( ++pCan->asynIgnorCounter[canNod] > MAX_ASYNCOUNTS ) pCan->asynIgnorCounter[canNod] = 0;
      printf("Async call ignore SDO segment flag! line: %d; preq addr: 0x%x; NodeId: %d; sdoRequstTime: %d; sdoReceiveTime: %d; pReq->devReadyTime: %d; pReq->requestTime: %d; pReq->responseTime: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pCan->sdoRequstTime[canNod], pReq->devParameters[sdoMsgTiStIndex], pReq->devReadyTime, pReq->requestTime, pReq->responseTime, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
      cdilog("Async call ignore SDO segment flag! line: %d; preq addr: 0x%x; NodeId: %d; sdoRequstTime: %d; sdoReceiveTime: %d; pReq->devReadyTime: %d; pReq->requestTime: %d; pReq->responseTime: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pCan->sdoRequstTime[canNod], pReq->devParameters[sdoMsgTiStIndex], pReq->devReadyTime, pReq->requestTime, pReq->responseTime, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
//      printf("Async call ignore SDO segment flag! line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
//      cdilog("Async call ignore SDO segment flag! line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
      return 1;
    }
    else
    {
      cdiPostSemaphore(&pCan->node_SemId);
      return 0;
    }
  }
  // for sync call, wait mximum WAIT_SDO_TOTOL_MS!
  while (pCan->segmentFlag[canNod])
  {
    wait_time.tv_sec = 0;
    wait_time.tv_nsec = WAIT_SDO_MS; 
    cdi_nanosleep( &wait_time, NULL );
    if ( getTimeMilliseconds() - pReq->devReadyTime > WAIT_SDO_TOTOL_MS ) break;
  }
  if ( pCan->segmentFlag[canNod] )
  {
    printf("Sync call ignore SDO segment flag! line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
    cdilog("Sync call ignore SDO segment flag! line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
  }
  return 1;
}
/****************************************************************************/
CDI_CAN_EXPORT int canLoadRequest(CdiRequestInfoBlk *pReq, int taskIndex)
{ int ln, i, canNod;
  CAN_LINE *pCan;

  if ( canLine[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1].sdoInterval == 0 ) return 0;
  if ( pReq->devParameters[msgType] != sdoMsg || pReq->accessFlag != cdiReadFlag ) return 0;
  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
  pCan = &canLine[ln];
  canNod = pReq->pDev->cdiAddr[caAddrIndex];
  for(i=0; i<pCan->numSdoBuf; i++)
    if ( canNod == pCan->sdoRevMsg[i].id && pReq->devParameters[canOpenObj] == pCan->sdoRevMsg[i].index && pReq->devParameters[canOpenSubIndex] == pCan->sdoRevMsg[i].subIndex ) break;
  if ( i == pCan->numSdoBuf )
  {
    if ( pCan->numSdoBuf < MAX_SDO_BUFFER )
    {
      pCan->sdoRevMsg[i].id = canNod;
      pCan->sdoRevMsg[i].index = pReq->devParameters[canOpenObj];
      pCan->sdoRevMsg[i].subIndex = pReq->devParameters[canOpenSubIndex];
      pCan->sdoRevMsg[i].timeStemp = 0;
      pCan->sdoRevMsg[i].dataLength = 0;
      pCan->numSdoBuf++;
      cdilog("num SDO Buffer = %d; line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; taskIndex: %d; SDO Indexes: 0x%x, 0x%x\n", pCan->numSdoBuf, ln+1, pReq, canNod, pReq->active, taskIndex, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
    }
    else
    {
      i = -1;
      cdilog("SDO Buffer full! line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; taskIndex: %d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pReq->active, taskIndex, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
      printf("SDO Buffer full! line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; taskIndex: %d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pReq->active, taskIndex, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
    }
  }
  pReq->devParameters[sdoMsgIndex] = i+1;
  return 0;
}
/****************************************************************************/
int pdoRead(CdiRequestInfoBlk *pReq)
{ int pdoTyp, retCode;
  unsigned int datalen;
  PDO_MSG_BUF *pdoBuf;
  char localData[8];

  pdoTyp = pReq->devParameters[msgType] - pdo1Msg;
  pdoBuf = &canLine[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1].pdoRevMsg[pdoTyp][pReq->pDev->cdiAddr[caAddrIndex]];
  cdiMutexLock(&pdoBuf->pdoMutex);
  retCode = 0;
  if ( abs(getTimeMilliseconds() - pdoBuf->timeStemp) < pReq->timeout )
  {
	datalen = pdoBuf->dataLength;
	if ( pReq->devParameters[canOpenObj] > 0 && pReq->devParameters[canOpenObj] < (int)datalen )
	{
	  datalen -= pReq->devParameters[canOpenObj];
	  memcpy(localData, &pdoBuf->pdoData[pReq->devParameters[canOpenObj]], datalen);
	}
	else memcpy(localData, pdoBuf->pdoData, datalen);
	checkReturnCondition(pReq, pReq->pInData, localData, datalen, 0);
	if ( pReq->dataReturnCondition != 0 )
	{
	  memcpy( pReq->pInData, localData, datalen > pReq->inLength ? pReq->inLength : datalen);
	  pReq->active = finishRequest;
	  retCode = 1;
	}
  }
  cdiMutexUnlock(&pdoBuf->pdoMutex);
  return retCode;
}
/****************************************************************************/
/* send request! input active flag = transRequest or transRequest2nd */
/* after, set active flag to finishRequest, recvRequest, or recvRequest2nd */
/****************************************************************************/
CDI_CAN_EXPORT void execCanbus(CdiRequestInfoBlk *pReq)
{ int wFlag;
  struct timespec wait_time;
  int diff, canNod, ln;
  char localData[4];
  CAN_LINE *pCan;
  static int insideFlag = 0;
  static int lastNode = 0;

  if ( pReq->accessFlag > cdiWriteReadFlag )
  {
    pReq->errorCode = not_implemented;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->active = finishRequest; /* finish request */
    return;
  }
  insideFlag++;
  if ( insideFlag > 1 )
  {
    //cdilog("canExec Inside Flag = %d  Call cdiExecRequest: activeFlag: %d; cancelFlag: %d; requestNumber: 0x%x; node = %d; lastNode = %d; preq addr: 0x%x\n", insideFlag, pReq->active, pReq->cancelFlag, pReq->requestNumber, pReq->pDev->cdiAddr[caAddrIndex], lastNode, pReq);
  }

  canNod = pReq->pDev->cdiAddr[caAddrIndex];
  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
  pCan = &canLine[ln];
  // 2013-8
  if ( pReq->devParameters[sdoMsgIndex] > 0 && pReq->active == transRequest && pReq->devParameters[msgType] == sdoMsg && pReq->accessFlag == cdiReadFlag &&
      abs(getTimeMilliseconds() - pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].timeStemp) < abs(pCan->sdoInterval) )
  {
    if ( pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength <= 4 )
    {
      memcpy( localData, pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].sdoData, pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength);
    	checkReturnCondition(pReq, pReq->pInData, localData, pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength, 0);
    }
    else pReq->dataReturnCondition = DATA_PASS_FILTER;
	  if ( pReq->dataReturnCondition != 0 )
  	{
      if ( pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength <= 4 )
        memcpy( pReq->pInData, localData, pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength > pReq->inLength ? pReq->inLength : pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength);
      else
        memcpy( pReq->pInData, pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].sdoData, pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength > pReq->inLength ? pReq->inLength : pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength);
      pReq->devParameters[sdoMsgCopyTiStIndex] = pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].timeStemp;
      pReq->active = finishRequest;
      //cdilog("use sdo buffer(%d) copy for node: %d(index: 0x%x: subindex: 0x%x);  line: %d\n", pReq->devParameters[sdoMsgIndex], canNod, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex], ln+1);
      cdiPostSemaphore(&pCan->node_SemId);
      insideFlag = 0;
      return;
	  }
  }
  // 2013-8
//  if ( pReq->active == transRequest && pReq->devParameters[msgType] == sdoMsg && pReq->accessFlag == cdiReadFlag &&
//    canNod == 32 && pReq->devParameters[canOpenObj] == 0x3004 && pReq->devParameters[canOpenSubIndex] == 0x0 )
//        cdilog("read sdo: time= %d; sdo buffer(%d) for node: %d(index: 0x%x: subindex: 0x%x)\n", getTimeMilliseconds(), pReq->devParameters[sdoMsgIndex], canNod, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);

  if ( pReq->devParameters[msgType] == sdoMsg )
  if ( pCan->sdoNoBlockFlag )	// 2009-9
  {
    if ( pCan->mutexFlag[canNod] == 0 )
    {
      cdiMutexInit(&pCan->sdoMutex[canNod], 1000 + canNod);
      pCan->mutexFlag[canNod] = 1;
      cdilog("create mutex for node: %d;  line: %d\n", canNod, ln+1);
    }
    cdiMutexLock(&pCan->sdoMutex[canNod]);

	diff = abs(getTimeMilliseconds() - pCan->sdoRequstTime[canNod]);
	if ( diff < WAIT_SDO_NOBLOCK )
	{
		wait_time.tv_sec = 0;
	    wait_time.tv_nsec = (WAIT_SDO_NOBLOCK - diff) * 1000000;	// ns
		cdi_nanosleep( &wait_time, NULL );
	}
  }
  else
  {
    if ( pCan->mutexFlag[canNod] == 0 )
    {
      cdiMutexInit(&pCan->sdoMutex[canNod], 1000 + canNod);
      pCan->mutexFlag[canNod] = 1;
      cdilog("create mutex for node: %d;  line: %d\n", canNod, ln+1);
    }
    cdiMutexLock(&pCan->sdoMutex[canNod]);
//    cnt=0;
//    while (pCan->segmentFlag[canNod])
//    {
//      wait_time.tv_sec = 0;
//      wait_time.tv_nsec = WAIT_SDO_MS; 
//      nanosleep( &wait_time, NULL );
//      if ( cnt++ == NUMBER_WAIT_SDO) break; // 200 ms
//    }
    if ( pCan->segmentFlag[canNod] )
    {
      if ( abs(getTimeMilliseconds() - pCan->sdoRequstTime[canNod]) < 2000 )
      {        
        cdilog("SDO segment flag busy(return)! line: %d; preq addr: 0x%x; NodeId: %d; sdoRequstTime: %d; sdoReceiveTime: %d; pReq->devReadyTime: %d; pReq->requestTime: %d; pReq->responseTime: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pCan->sdoRequstTime[canNod], pReq->devParameters[sdoMsgTiStIndex], pReq->devReadyTime, pReq->requestTime, pReq->responseTime, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
//        cdilog("SDO segment flag busy(return)! line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, canNod, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
        pReq->errorCode = device_busy;
        if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
        pReq->active = finishRequest; /* finish request */
        cdiMutexUnlock(&pCan->sdoMutex[canNod]);
        cdiPostSemaphore(&pCan->node_SemId);
        insideFlag = 0;
        return;
      }
      cdilog("SDO segment flag not cleared(continue)! line: %d; preq addr: 0x%x, NodeId: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x; lasSDOtTime: %d; pReq; 0x%x\n", ln+1, pReq, canNod, pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex], pCan->sdoRequstTime[canNod], pCan->reqBufPointer[canNod]);
    }
    pReq->devParameters[transState] = 0;
    pCan->sdoRequstTime[canNod] = getTimeMilliseconds();
    pCan->segmentFlag[canNod] = 1;
    pCan->reqBufPointer[canNod] = pReq;
    cdiMutexUnlock(&pCan->sdoMutex[canNod]);
  }

  if ( pReq->active == transRequest ) /* start first transfere */
  {
	if ( pReq->devParameters[msgType] >= pdo1Msg && pReq->devParameters[msgType] <= nmtMsg && pReq->accessFlag == cdiReadFlag && pReq->devParameters[canAutoReceive] ) if ( pdoRead(pReq) ) goto goExit;
//	if ( pReq->devParameters[msgType] >= pdo1Msg && pReq->devParameters[msgType] <= pdo4Msg && pReq->accessFlag == cdiReadFlag && pReq->devParameters[canAutoReceive] ) if ( pdoRead(pReq) ) goto goExit;
    pReq->active = recvRequest; /* afterwards goto receive */
    /* auto receive, without request! */
    // change to avoid goto exit with different CanOpenObj for read and write acess of SDO  2013-06
//    if ( (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) && pReq->devParameters[canAutoReceive] ) goto goExit;
    if ( (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) && pReq->devParameters[canAutoReceive] && (pReq->devParameters[msgType] != sdoMsg || pReq->devParameters[canBroadcast] == 0)) goto goExit;
    wFlag = (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag) ? 1 : 0;
  }
  else
  {
    pReq->active = recvRequest2nd; /* goto 2ndReceive */
    if ( pReq->accessFlag == cdiWriteReadFlag && pReq->devParameters[canAutoReceive] ) goto goExit;
    wFlag = pReq->accessFlag == cdiReadWriteFlag ? 1 : 0;
  }

  if ( sendCdiCanMessage(pReq, wFlag) < 0 ) goto goExit;

  /* without message back */
  if ( wFlag && (pReq->devParameters[canOpenId] == PDO_TRANSMIT || pReq->devParameters[canOpenId] == PDO2_TRANSMIT || pReq->devParameters[canOpenId] == PDO3_TRANSMIT || pReq->devParameters[canOpenId] == PDO4_TRANSMIT || pReq->devParameters[canOpenId] == NMT_COMMANDS) )
  {
    // last write
    if (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiReadWriteFlag) pReq->active = finishRequest;
    else
    { /* cdiWriteReadFlag */
      pReq->active = recvRequest2nd; /* goto 2ndReceive */
      if ( pReq->devParameters[canAutoReceive] == 0 ) sendCdiCanMessage(pReq, 0);
    }
  }
//  if ( wFlag && (pReq->devParameters[canOpenId] == PDO_TRANSMIT || pReq->devParameters[canOpenId] == PDO2_TRANSMIT || pReq->devParameters[canOpenId] == NMT_COMMANDS) ) pReq->active = finishRequest;

  if ( NGdebug ) cdilog("active = %d; preq addr: 0x%x\n", pReq->active, pReq);

goExit:
  if ( pReq->devParameters[msgType] == sdoMsg && pCan->sdoNoBlockFlag )	// 2009-9
    cdiMutexUnlock(&pCan->sdoMutex[canNod]);

  insideFlag = 0;
  if ( pReq->devParameters[msgType] == sdoMsg ) cdiPostSemaphore(&pCan->node_SemId);
  return;
}

/****************************************/
/*	can receive thread					*/
/****************************************/
#ifdef CDI_CAN_LINUX
void *GetCanDataThread(void *param)
#else
DWORD WINAPI GetCanDataThread(void *param)
#endif
{ int i;
  unsigned int len, errCode, ret;
  char *pErrStr;
  CAN_LINE *pCan = (CAN_LINE *)param;
  struct timespec wait_time;
#if defined(ESDCAN_H_) && defined(CDI_CAN_LINUX) && defined(ESDCAN_FIXED)
//changed on 2016-11
  ret = canDrvOpen(pCan->lineNumber, pCan->devPath, pCan->canRate, pCan->net, pCan->mode, pCan->txqueuesize, pCan->rxqueuesize, pCan->txtimeout, pCan->rxtimeout, &(pCan->readfd), &errCode);
//  ret = canDrvOpen(pCan->net, pCan->devPath, pCan->canRate, pCan->net, pCan->mode, pCan->txqueuesize, pCan->rxqueuesize, pCan->txtimeout, pCan->rxtimeout, &(pCan->readfd), &errCode);
  if (ret)
  {
    cdilog("GetCanDataThread canOpen() failed with error: %d; ESD can line: %d\n",errCode, pCan->net + 1);
    printf("GetCanDataThread canOpen() failed with error: %d; ESD can line: %d\n",errCode, pCan->net + 1);
	exit(0);
  }
#endif
  do
  {
//    getCanTelegram(pCan);
//    for ( pCan->recCnt = 0; pCan->recCnt < pCan->recLength; pCan->recCnt++)
//    {
/*      pCan->rcvPmsg = &pCan->cmsg[pCan->recCnt];
      pCan->retId = (short)pCan->rcvPmsg->id;
      pCan->rtr = (pCan->rcvPmsg->len >> 4) & 0x1;
      pCan->length = pCan->rcvPmsg->len & 0xf;

      memcpy(pCan->retData, pCan->rcvPmsg->data, pCan->length);
      cpReceiveTelegram(pCan);
*/
	  if ( pCan->statusOK != 0 )
	  {
		wait_time.tv_sec = 1;
		wait_time.tv_nsec = 0;
		cdi_nanosleep( &wait_time, NULL );
		continue;
	  }

#if defined(ESDCAN_H_) && defined(CDI_CAN_LINUX) && defined(ESDCAN_FIXED)
      readCanTelegram(pCan->lineNumber, pCan->readfd, &(pCan->retId), &(pCan->rtr), &(pCan->length), pCan->retData);
#else
      readCanTelegram(pCan->lineNumber, pCan->fd, &(pCan->retId), &(pCan->rtr), &(pCan->length), pCan->retData);
#endif
      pCan->remoteNode = pCan->retId & 0x7f;
      pCan->remoteCommand = pCan->retId & 0xff80;


//2013 debug
//      if ( pCan->remoteCommand == SDO_RECEIVE && pCan->remoteNode == 32 )
//         cdilog("Receive SDO: line: %d; NodeId: %d; SDO Indexes: 0x%x, 0x%x\n", pCan->lineNumber, pCan->remoteNode, *(short *)&pCan->retData[1], pCan->retData[3]);

	  len = 0;
      if ( pCan->remoteCommand == SDO_RECEIVE ) pCan->numRevMsg[sdoMsg][pCan->remoteNode]++;
      else if ( pCan->remoteCommand == PDO_DI_READ ) {len = 1; pCan->numRevMsg[pdo1Msg][pCan->remoteNode]++;}
      else if ( pCan->remoteCommand == PDO_AI_READ ) {len = 2; pCan->numRevMsg[pdo2Msg][pCan->remoteNode]++;}
      else if ( pCan->remoteCommand == PDO3_READ ) {len = 3; pCan->numRevMsg[pdo3Msg][pCan->remoteNode]++;}
      else if ( pCan->remoteCommand == PDO4_READ ) {len = 4; pCan->numRevMsg[pdo4Msg][pCan->remoteNode]++;}
      else if ( pCan->remoteCommand == NMT_ERROS ) {len = 5; pCan->numRevMsg[nmtMsg][pCan->remoteNode]++;}
      else if ( pCan->remoteCommand == EMERGENCE_MESSAGE ) pCan->numRevMsg[emergencyMsg][pCan->remoteNode]++;
      else pCan->numRevMsg[notdefinedMsg][pCan->remoteNode]++;

//	printf("node: %d; command: 0x%x; number of messages: %d\n",pCan->remoteNode, pCan->remoteCommand, pCan->numRevMsg[nmtMsg][pCan->remoteNode]);

	  if ( len )  /* store pdo message	*/
	  {
		len--;
	    cdiMutexLock(&pCan->pdoRevMsg[len][pCan->remoteNode].pdoMutex);
		pCan->pdoRevMsg[len][pCan->remoteNode].timeStemp = getTimeMilliseconds();
		pCan->pdoRevMsg[len][pCan->remoteNode].pdoType = len;
		pCan->pdoRevMsg[len][pCan->remoteNode].dataLength = pCan->length;
		memcpy(pCan->pdoRevMsg[len][pCan->remoteNode].pdoData, pCan->retData, pCan->length);
	    cdiMutexUnlock(&pCan->pdoRevMsg[len][pCan->remoteNode].pdoMutex);
	  }

      if ( scanFlag && (len=strlen(scanBuffer)) < scanBufferSize - 30)
      {
        if ( scanFlag >= startremote && scanFlag <= resetcommunication || (pCan->remoteNode == scanNodId &&
          ( ((scanFlag == writepdo1 || scanFlag == readpdo1) && pCan->remoteCommand == PDO_DI_READ) ||
            ((scanFlag == writenmt || scanFlag == readnmt) && (pCan->remoteCommand == NMT_ERROS || pCan->remoteCommand == EMERGENCE_MESSAGE) ) ||
            ((scanFlag == writesdo || scanFlag == readsdo) && pCan->remoteCommand == SDO_RECEIVE) ||
            ((scanFlag == writepdo2 || scanFlag == readpdo2) && pCan->remoteCommand == PDO_AI_READ) ||
            ((scanFlag == writepdo3 || scanFlag == readpdo3) && pCan->remoteCommand == PDO3_READ) ||
            ((scanFlag == writepdo4 || scanFlag == readpdo4) && pCan->remoteCommand == PDO4_READ) ) ) )
        {
          snprintf(&scanBuffer[len],scanBufferSize-len,"Node: %d, MsgID: 0x%x, rtr: %d, Length: %d, ",pCan->remoteNode, pCan->remoteCommand, pCan->rtr, pCan->length);
          for(i=0; i<pCan->length; i++)
          {
            if ( (len = strlen(scanBuffer)) < scanBufferSize - 4)
              snprintf(&scanBuffer[len],scanBufferSize-len,"%2x ", ((int)pCan->retData[i] & 0xff));
          }
          if ((scanFlag == writesdo || scanFlag == readsdo) && pCan->remoteCommand == SDO_RECEIVE && (pCan->retData[0] & 0xf0) == SDO_ABORT )// sdo abort Domain Transfer
          {
            errCode = *(unsigned int *)&pCan->retData[4];
            if ( errCode == 0x05040001 ) pErrStr = "Error: illegal Function";
            else if ( errCode == 0x06010002 ) pErrStr = "Error: Access unsupported(Read Only)";
            else if ( errCode == 0x06020000 ) pErrStr = "Error: Object(index) nonexistent";
            else if ( errCode == 0x06070010 ) pErrStr = "Error: Data Type(length) Conflict";
            else if ( errCode == 0x06090011 ) pErrStr = "Error: Subindex nonexistent";
            else pErrStr = "Error: SDO error not defined";
            if ( (len = strlen(scanBuffer)) < scanBufferSize - 20) snprintf(&scanBuffer[len],scanBufferSize-len,"%s(0x%x)",pErrStr,errCode);
          }
          if ( (len = strlen(scanBuffer)) < scanBufferSize - 2) snprintf(&scanBuffer[len],scanBufferSize-len,"\n");
        }
      }

      if ( pCan->remoteCommand != NMT_ERROS )
      {
        if ( canMsgPrint )
        {
          printf("Master ID %x, rtr: %d; length: %d : ", pCan->retId, pCan->rtr, pCan->length);
          for(i=0; i<pCan->length; i++) printf("%2x ", ((int)pCan->retData[i] & 0xff));
          printf("\n");
        }
        if ( pCan->receiveCanMessage ) pCan->receiveCanMessage(pCan->cdiLine, pCan->retData, pCan->length, pCan->retId, pCan->rtr, pCan->remoteNode, pCan->remoteCommand);
      }
      else
      {
        /* guarding or heartbeat message: powerOn or preOperation mode */
        if ( (pCan->retData[0] == 0x0 || pCan->retData[0] == 0x7f) && pCan->length == 1 )
        {
          printf("Start Operation on Node ID 0x%x, line number: %d; retData[0] = 0x%x\n", pCan->remoteNode, pCan->lineNumber + 1, pCan->retData[0]);
          cdilog("Start Operation on Node ID 0x%x, line number: %d; retData[0] = 0x%x\n", pCan->remoteNode, pCan->lineNumber + 1, pCan->retData[0]);
          i = NMT_START_REMOTE;
//        	if ( ret = canMsgSend(pCan->lineNumber + 1, 0, NMT_COMMANDS, 1, (char *)&i, 0, 0, 0, 0)  )
		  if ( ret = canMsgSend(pCan->lineNumber + 1, pCan->remoteNode, NMT_COMMANDS, 1, (char *)&i, 0, 0, 0, 0)  )	// start single one dose not work!! 2010
          {
            cdilog("Start Operation: canMsgSend error(0x%x) on Node ID 0x%x, line number: %d\n", ret, pCan->remoteNode, pCan->lineNumber + 1);
          }
        }
        // 20.12.10 permits read of nmt message with type 5 (HGW)
        if ( pCan->receiveCanMessage ) pCan->receiveCanMessage(pCan->cdiLine, pCan->retData, pCan->length, pCan->retId, pCan->rtr, pCan->remoteNode, pCan->remoteCommand);
      }
//    }
  } while(1);
}

/****************************************/
/*	CAN send API						*/
/*	return -1: line number wrong		*/
/*	return -2: not init yet				*/
/*	return =0: send error				*/
/****************************************/
int canMsgSend(int line, short nodeId, short canOpenId, short length, char *data, short ObjNum, short subIndex, short wFlag, short sdoSegmentCmd)
{ CAN_LINE *pCan;
  int rtr, ret;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd == 0 ) return -file_not_open;
//  if ( pCan->fd < 0 ) return -file_not_open;

  if ( nodeId < 0 ) nodeId = 0;

//  printf("***canMsgSend ==> line: %d; nodeID: %d; CanOPENID: %d\n",line, nodeId, canOpenId);

  switch (canOpenId)
  {
  	case NMT_COMMANDS:
	    canOpenId = NMT_COMMANDS;
	    rtr = 0;
	    length = 2;
	    pCan->sendData[0] = data[0];	// only one byte data for EXEC_REMOTE
	    pCan->sendData[1] = (unsigned char)nodeId;
	    break;
	  case SDO_TRANSMIT:
      if ( sdoSegmentCmd || ( ObjNum == 0 && subIndex == 0 && wFlag ) ) pCan->sendData[0] = (char)sdoSegmentCmd;
	    else pCan->sendData[0] = wFlag ? SDO_WRITE : SDO_READ;
      if ( ObjNum == 0 && subIndex == 0 && wFlag )
        memcpy(&pCan->sendData[1],data,length);
      else
      {
  	    *(short *)&pCan->sendData[1] = ObjNum;	// object 6300; sub index 1
	      pCan->sendData[3] = (unsigned char)subIndex;
	      if ( wFlag ) memcpy(&pCan->sendData[4],data,length);
//      if ( wFlag && length == 4 ) printf("can save data: %x\n",*(int *)&FastMsg.data[6]);
      }
	    length = 8;
	    rtr = 0;
      break;
  	case PDO_TRANSMIT:
  	case PDO2_TRANSMIT:
  	case PDO3_TRANSMIT:
  	case PDO4_TRANSMIT:
	    memcpy(&pCan->sendData[0],data,length);
	    rtr=0;
	    break;
	case PDO_DI_READ:
	case PDO_AI_READ:
	case PDO3_READ:
	case PDO4_READ:
	    length = 0;
	    rtr = 1;
	    break;
	  default:
	    cdilog("canSendAPI: ****Wrong CanOpenID: 0x%x; line: %d\n",canOpenId, line+1);
	    return -3;
  }
  pCan->sendId = canOpenId;
  if ( canOpenId != NMT_COMMANDS) pCan->sendId += nodeId;
//  pCan->sendId = canOpenId + nodeId;				// can message used in can send
  pCan->sendRTR = rtr;
  pCan->sendLength = (char)length;
  pCan->xmtLength = 1;

  if ( NGdebug ) cdilog("CAN send id: 0x%x; len: 0x%x; rtr: %d; data: 0x%x 0x%x 0x%x 0x%x \n",pCan->sendId, pCan->sendLength, pCan->sendRTR, pCan->sendData[0], pCan->sendData[1], pCan->sendData[2], pCan->sendData[3]);

//  pCan->xmtStatus = canWrite(pCan->fd, &pCan->xmtMsg, &pCan->xmtLength, NULL);
//  return outCanTelegram(pCan);
  ret = sendCanTelegram(line, pCan->fd, pCan->sendId, pCan->sendRTR, pCan->sendLength, pCan->sendData, &(pCan->xmtStatus));
  if ( ret == -1 )
    cdilog("ESD CAN send Error! return code: 0x%x\n",pCan->xmtStatus); 
  else if ( ret == -2 )
    cdilog("PEAK CAN send Error! return code: 0x%x\n",pCan->xmtStatus); 
  if ( ret )
  {
    pCan->numSendErrors++;
    return -hardware_error;
  }
  return 0;
}

/****************************************/
void doneRequest(CdiRequestInfoBlk *pReq, int errorCode)
{ int canNod = pReq->pDev->cdiAddr[caAddrIndex];
  int ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
  CAN_LINE *pCan = &canLine[ln];
  if ( errorCode )
  {
    pReq->errorCode = errorCode;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
  }
  pReq->devParameters[transState] = 0;
//  if (pReq->devParameters[msgType] == sdoMsg)
  if ( pReq->devParameters[msgType] == sdoMsg && pCan->sdoNoBlockFlag == 0 )	// 2009-9
  {
    cdiMutexLock(&pCan->sdoMutex[canNod]);
    pCan->segmentFlag[canNod] = 0;
    pCan->reqBufPointer[canNod] = NULL;
    cdiMutexUnlock(&pCan->sdoMutex[canNod]);
  }
//  cdiWaitSemaphore(&pReq->exec_SemId,0);
  cdiExecRequest(pReq, readTaskIndex);
//  cdiPostSemaphore(&pReq->exec_SemId);
}
/****************************************/
/*	init can line						*/
/****************************************/
/**************************************/
void canMessage(int cdiLine, char *data, int leng, int retId, int rtr, int remoteNode, int remoteCommand)
{ 
  /* CdiLineInfoBlk *pLine = busHandle[cdiLine]; */
  CdiLineInfoBlk *busHandle;
  CdiRequestInfoBlk *pReq;
  short readFlag, returnObj, index, sendObj, sendIndex;
  char *str;
  unsigned int datalen;
  char localData[8];
  SDO_PROTOCOL  *prot;
  SDO_SEG_PROTOCOL *protSegment;
  int ret, ln;
  CAN_LINE *pCan;

  /* find out device */
  if ((busHandle=cdiGetBusHandle(cdiLine)) == NULL)
  {
    cdilog("canMessage: Wrong cdi_line: %d for canMessage() call\n",cdiLine);
    return;
  }
  
  if (NGdebug) 
  {
    cdilog("CanMessage call Back: cdi line: %d, can line number: %d, data: %s, length: %d, retId: %d, rtr: %d\n",
            cdiLine, busHandle->busLine, data, leng, retId, rtr);
    cdilog("Number of active request: %d\n",busHandle->numberActiveRequest);
  }

  for ( busHandle->readCounter=0; busHandle->readCounter < busHandle->numberActiveRequest; busHandle->readCounter++ )
  {
    if ( NGdebug ) cdilog("enter loop ! Number of active request: %d; active flag = %d; preq addr: 0x%x\n",busHandle->numberActiveRequest, busHandle->pReqList[ busHandle->readCounter ].active, &busHandle->pReqList[ busHandle->readCounter ]);

    if ( (busHandle->pReqList[ busHandle->readCounter ].active != recvRequest &&
      busHandle->pReqList[ busHandle->readCounter ].active != recvRequest2nd) ||
      busHandle->pReqList[ busHandle->readCounter ].pDev->cdiAddr[caAddrIndex] != remoteNode ) continue;


    pReq = &busHandle->pReqList[ busHandle->readCounter ];
	ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
	pCan = &canLine[ln];
    
    cdiWaitSemaphore(&pReq->exec_SemId,0);
    if ( (pReq->active != recvRequest && pReq->active != recvRequest2nd) || pReq->pDev->cdiAddr[caAddrIndex] != remoteNode )
    {
      cdiPostSemaphore(&pReq->exec_SemId);
      continue;
    }
    if ( pReq->active == recvRequest ) /* first part */
      readFlag = (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) ? 1 : 0;
    else /* second part! */
      readFlag = (pReq->accessFlag == cdiWriteReadFlag) ? 1 : 0;
    returnObj = *(short *)&data[1];

    datalen = leng;
    str = data;
    if ( pReq->devParameters[canOpenId] == SDO_TRANSMIT ) /* check messagesType */
    {
      if ( remoteCommand != SDO_RECEIVE ) goto clrSem;  //continue;
      index = getSDOindex(pReq);
      sendObj = index ? pReq->devParameters[canBroadcast] : pReq->devParameters[canOpenObj];
      sendIndex = index ?  pReq->devParameters[canAutoReceive] : pReq->devParameters[canOpenSubIndex];
      if ( pCan->sdoNoBlockFlag == 0 )	// 2009-9
	  {
      if ( pCan->segmentFlag[pReq->pDev->cdiAddr[caAddrIndex]] == 0 )
      {
        cdilog("Receive SDO from slave without SDO Transfere flag!! line: %d; preq addr: 0x%x; NodeId: %d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pReq, remoteNode, returnObj, data[3]);
        goto clrSem;
      }
      else if ( pCan->reqBufPointer[pReq->pDev->cdiAddr[caAddrIndex]] != pReq )
      {
        cdilog("Receive SDO from slave, do not belong signed pReq Buffer!! line: %d; signed pReq: 0x%x; preq addr: 0x%x; NodeId: %d; SDO Indexes: 0x%x, 0x%x\n", ln+1, pCan->reqBufPointer[pReq->pDev->cdiAddr[caAddrIndex]], pReq, remoteNode, returnObj, data[3]);
        goto clrSem;
      }
	  }
      if ( pReq->devParameters[transState] == 0 )
        if ( sendObj != returnObj || (char)sendIndex != data[3] ) goto clrSem;  //continue;
      prot = (SDO_PROTOCOL *)&data[0];
      protSegment = (SDO_SEG_PROTOCOL *)&data[0];
      if ( (data[0] & 0xf0) == SDO_ABORT ) // sdo abort Domain Transfer
      {
        if ( pReq->devParameters[transState] &&
          (sendObj != returnObj || (char)sendIndex != data[3]) ) goto clrSem;  //continue;
        pReq->deviceError = *(int *)&data[4];
        if ( pReq->deviceError == 0x05040001 ) datalen = illegal_mode;
        else if ( pReq->deviceError == 0x06010002 ) datalen = illegal_operator;
        else if ( pReq->deviceError == 0x06020000 ) datalen = invalid_index;
        else if ( pReq->deviceError == 0x06070010 ) datalen = device_not_connected;
        else if ( pReq->deviceError == 0x06090011 ) datalen = illegal_device;
        else datalen = not_defined;
        cdilog("Slave SDO abort command: 0x%x; NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",pReq->deviceError,remoteNode, returnObj, (int)data[3], pReq);
        doneRequest(pReq, datalen);
        goto clrSem;  //continue;
      }
      else if ( readFlag ) /* SDO read */
      {
        if ( pReq->devParameters[transState] == 0 )
        {
          if ( prot->s.scs != 2 ) goto clrSem;  //continue;           //if ( (data[0] & 0xf0) != 0x40 ) continue;
//          if ( prot->s.e == 0 && prot->s.s == 1 )     // confirm SDO init upload
          if ( prot->s.e == 0 && prot->s.s == 1 && pCan->sdoNoBlockFlag == 0 )	// 2009-9 // confirm SDO init upload
          {
            datalen = *(int *)&data[4];
            if ( datalen == 0 )
            {
              datalen = 0x06040043;   // abort general parameter incompatibility reason
              cdilog("Active SDO abort during reading: general parameter incompatibility(0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",datalen,remoteNode, returnObj, (int)data[3], pReq);
              ret = canMsgSend(busHandle->busLine, (short)remoteNode, SDO_TRANSMIT, 4, (char *)&datalen, sendObj, sendIndex, 1, SDO_ABORT);
              doneRequest(pReq, invalid_value);
              goto clrSem;  //continue;
            }
            pReq->devParameters[downloadFlag] = 0;    // init segment transfer parameters
            pReq->devParameters[dataIndex] = 0;
            pReq->devParameters[dataOverflow] = 0;
            pReq->devParameters[transState] = datalen;
            ret = canMsgSend(busHandle->busLine, (short)remoteNode, SDO_TRANSMIT, 0, (char *)&datalen, sendObj, sendIndex, 0, SDO_READ_SEGMENT);
            if ( ret ) 
            {
              cdilog("SDO read:init segment transfer parameter, Can send error: (0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",-ret,remoteNode, returnObj, (int)data[3], pReq);
              doneRequest(pReq, -ret);
            }
            goto clrSem;  //continue;
          }
          else    // normal read data
		  {
            datalen = (prot->s.e == 1 && prot->s.s == 1) ? (4 - prot->s.n) : 4;
//			printf("Receive SDO coming: length: %d; NodeId: %d; SDO Indexes: 0x%x, 0x%x; responseTime: %d; requestTime: %d; now: %d\n",datalen,remoteNode, returnObj, data[3], pReq->responseTime, pReq->requestTime, getTimeMilliseconds());
		  }
        }
        else
        {
          if ( protSegment->s.scs != 0 ) goto clrSem;  //continue;
          if ( protSegment->s.t != (unsigned)pReq->devParameters[downloadFlag] ) // check toggle bit
          {
            datalen = 0x05030000;   // abort Toggle bit reason
            cdilog("Active SDO abort during reading: Toggle bit wrong(0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",datalen,remoteNode, sendObj, sendIndex, pReq);
            ret = canMsgSend(busHandle->busLine, (short)remoteNode, SDO_TRANSMIT, 4, (char *)&datalen, sendObj, sendIndex, 1, SDO_ABORT);
            doneRequest(pReq, invalid_value);
            goto clrSem;  //continue;
          }
          datalen = 7 - protSegment->s.n;
          index = pReq->devParameters[dataIndex];
          memcpy( &((char *)pReq->pInData)[index], &str[1], datalen > (pReq->inLength - index) ? (pReq->inLength - index) : datalen);
          if (datalen > (pReq->inLength - index) ) pReq->devParameters[dataOverflow] = 1;
          pReq->devParameters[dataIndex] += datalen;
          ret = 0;
          if ( protSegment->s.c == 0 )    //  more segment to be loaded
          {
            if ( (unsigned)pReq->devParameters[dataIndex] >= pReq->inLength )   // finish upload
            {
              datalen = 0x05040005;   // abort out of memory
              cdilog("Active SDO abort during reading: Out of Memory(0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",datalen,remoteNode, sendObj, sendIndex, pReq);
              ret = canMsgSend(busHandle->busLine, (short)remoteNode, SDO_TRANSMIT, 4, (char *)&datalen, sendObj, sendIndex, 1, SDO_ABORT);
              if ( ret )
              {
                cdilog("SDO read:finish upload out of memory, Can send error: (0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",-ret,remoteNode, sendObj, sendIndex, pReq);
                ret = -ret;
              }
            }
            else        // continue transfer
            {
              pReq->devParameters[downloadFlag] = pReq->devParameters[downloadFlag] ? 0 : 1;
              datalen = SDO_READ_SEGMENT;
              if ( pReq->devParameters[downloadFlag] ) datalen += 0x10;   // add toggle bit
              ret = canMsgSend(busHandle->busLine, (short)remoteNode, SDO_TRANSMIT, 0, (char *)&datalen, sendObj, sendIndex, 0, (short)datalen);
              if ( ret ) 
              {
                cdilog("SDO read:continue transfer, Can send error: (0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",-ret,remoteNode, sendObj, sendIndex, pReq);
                doneRequest(pReq, -ret);
              }
              goto clrSem;  //continue;
            }
          }
          pReq->dataReturnCondition = DATA_PASS_FILTER; /* flag for callBack */
          if ( ret == 0 && pReq->devParameters[sdoMsgIndex] && pReq->devParameters[dataOverflow] == 0 && pReq->devParameters[dataIndex] < STORE_SDO_LENGTH) // 2013-8
          {
            pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength = pReq->devParameters[dataIndex];
            memcpy( pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].sdoData, pReq->pInData, pReq->devParameters[dataIndex]);
            pReq->devParameters[sdoMsgTiStIndex] = pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].timeStemp = getTimeMilliseconds();
          }
          doneRequest(pReq, ret);
          goto clrSem;  //continue;
        }
      }
      else   /* SDO write */
      {
        if ( pReq->devParameters[transState] == 0 )
        {
          if ( prot->s.scs != 3 ) goto clrSem;  //continue;          //if ( (data[0] & 0xf0) != 0x60 ) continue;
//          if ( pReq->outLength > 4 )                 // confirm SDO init download
          if ( pReq->outLength > 4 && pCan->sdoNoBlockFlag == 0 )	// 2009-9 // confirm SDO init download
          {
            pReq->devParameters[downloadFlag] = 0;
            pReq->devParameters[dataIndex] = pReq->outLength < 8 ? pReq->outLength : 7;
            pReq->devParameters[transState] = pReq->outLength;
            protSegment->s.scs = 0;
            protSegment->s.t = 0;
            protSegment->s.n = 7 - pReq->devParameters[dataIndex];
            protSegment->s.c = pReq->outLength < 8 ? 1 : 0;
            ret = canMsgSend(busHandle->busLine, (short)remoteNode, SDO_TRANSMIT, (short)pReq->devParameters[dataIndex], (char *)pReq->pOutData, 0, 0, 1, (short)protSegment->value);
            if ( ret ) 
            {
              cdilog("SDO write:confirm  init download, Can send error: (0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",-ret,remoteNode, returnObj, (int)data[3], pReq);
              doneRequest(pReq, -ret);
            }
            goto clrSem;  //continue;
          }
          else
		  {
			  datalen = leng - 4;      // normal write data
//			  printf("Write SDO coming: length: %d; NodeId: %d; SDO Indexes: 0x%x, 0x%x\n",datalen,remoteNode, returnObj, data[3]);
		  }
        }
        else
        {
          if ( protSegment->s.scs != 1 ) goto clrSem;  //continue;
          if ( protSegment->s.t != (unsigned)pReq->devParameters[downloadFlag] ) // check toggle bit
          {
            datalen = 0x05030000;   // abort Toggle bit reason
            cdilog("Active SDO abort during writing: Toggle bit wrong(0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",datalen,remoteNode, sendObj, sendIndex, pReq);
            ret = canMsgSend(busHandle->busLine, (short)remoteNode, SDO_TRANSMIT, 4, (char *)&datalen, sendObj, sendIndex, 1, SDO_ABORT);
            doneRequest(pReq, invalid_value);
            goto clrSem;  //continue;
          }
          if ( (unsigned)pReq->devParameters[dataIndex] < pReq->outLength )
          {
            pReq->devParameters[downloadFlag] = pReq->devParameters[downloadFlag] ? 0 : 1;
            datalen = (pReq->outLength - pReq->devParameters[dataIndex]) < 8 ? (pReq->outLength - pReq->devParameters[dataIndex]) : 7;
            protSegment->s.scs = 0;
            protSegment->s.t = pReq->devParameters[downloadFlag];
            protSegment->s.n = 7 - datalen;
            protSegment->s.c = (pReq->outLength - pReq->devParameters[dataIndex]) < 8 ? 1 : 0;
            ret = canMsgSend(busHandle->busLine, (short)remoteNode, SDO_TRANSMIT, (short)datalen, (char *)&((char *)pReq->pOutData)[pReq->devParameters[dataIndex]], 0, 0, 1, (short)protSegment->value);
            pReq->devParameters[dataIndex] += datalen;
            if ( ret ) 
            {
              cdilog("SDO write:continue send, Can send error: (0x%x); NodeId: %d; Obj Number: 0x%x; Obj Index: 0x%x; pReq: 0x%x",-ret,remoteNode, sendObj, sendIndex, pReq);
              doneRequest(pReq, -ret);
            }
            goto clrSem;  //continue;
          }
          doneRequest(pReq, 0);
          goto clrSem;  //continue;
        }
      }
      /* SDO data back.... */
//      datalen = leng - 4;
      str = &data[4];
    }
    else if ( pReq->devParameters[canOpenReadId] != notdefinedMsg && pReq->devParameters[canOpenReadId] != remoteCommand ) goto clrSem;  //continue; /* same OpenId? */
    else if ( readFlag == 0 ) /* for write command device should not answer! */
    {
      if (NGdebug) 
      {
        cdilog("canMessage Wrong: unexpected return message for write conmmand!");
        cdilog("CanOpenId: 0x%x; canLine: %d; module CanId: %d; module Name: %s",
                remoteCommand,busHandle->busLine,remoteNode,pReq->pDev->devName);
      }
      goto clrSem;  //continue;
    }
	else if ( rtr )	// rtr bit is on, discard it
	{
      if (NGdebug) 
      {
        cdilog("canMessage rtr bit ON, discard it:\n");
        cdilog("CanOpenId: 0x%x; canLine: %d; module CanId: %d; module Name: %s",
                remoteCommand,busHandle->busLine,remoteNode,pReq->pDev->devName);
	  }
      goto clrSem;  //continue;
	}
	if ( pReq->devParameters[msgType] >= pdo1Msg && pReq->devParameters[msgType] <= pdo4Msg && pReq->devParameters[canOpenObj] > 0 && pReq->devParameters[canOpenObj] < (int)datalen )
	{
		datalen -= pReq->devParameters[canOpenObj];
		memcpy(localData, &str[pReq->devParameters[canOpenObj]], datalen);
	}
	else
		memcpy(localData, str, datalen);
    if ( readFlag ) /* check return condition */
    {
      if (NGdebug) cdilog("CAN data check condition for readFlag canOpenID: 0x%x",remoteCommand);
      checkReturnCondition(pReq, pReq->pInData, localData, datalen, 0);
      if ( pReq->dataReturnCondition == 0 ) goto clrSem;  //continue; 
    }

    if (NGdebug) cdilog("can data return condition OK canOpenID: 0x%x",remoteCommand);
    
    if ( readFlag ) memcpy( pReq->pInData, localData, datalen > pReq->inLength ? pReq->inLength : datalen);

    if ( pReq->devParameters[canOpenId] == SDO_TRANSMIT && readFlag && pReq->devParameters[sdoMsgIndex] ) // 2013-8
    {
      pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].dataLength = datalen;
      memcpy( pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].sdoData, str, datalen);
      pReq->devParameters[sdoMsgTiStIndex] = pCan->sdoRevMsg[ pReq->devParameters[sdoMsgIndex]-1 ].timeStemp = getTimeMilliseconds();
    }
//    if (readFlag && pReq->pDev->cdiAddr[caAddrIndex] == 49 && pReq->devParameters[canOpenObj] == 0x3001 && pReq->devParameters[canOpenSubIndex] == 0 )
//      cdilog("read sdo! offset: %d; SDO Buffer = %d; line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; SDO Indexes: 0x%x, 0x%x\n",  sdoMsgIndex, pReq->devParameters[sdoMsgIndex], ln+1, pReq, pReq->pDev->cdiAddr[caAddrIndex], pReq->active, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);
//          cdilog("num SDO Buffer = %d; line: %d; preq addr: 0x%x; NodeId: %d; activeMode=%d; taskIndex: %d; SDO Indexes: 0x%x, 0x%x\n", pCan->numSdoBuf, ln+1, pReq, pReq->pDev->cdiAddr[caAddrIndex], pReq->active, taskIndex, pReq->devParameters[canOpenObj], pReq->devParameters[canOpenSubIndex]);

    doneRequest(pReq, 0);
clrSem:
    cdiPostSemaphore(&pReq->exec_SemId);
  }
}

#define CDI_PARAMETER_DELIMETER ":"
int cdiParseParamString(char *paramstr,int *parameters)
{
  char *s,*c=NULL;
  int i;
  if (paramstr == NULL || parameters == NULL) return argument_list_error;
  for (s=strtok_r(paramstr,CDI_PARAMETER_DELIMETER,&c), i=0; 
       i<NUMBER_EXEC_PARAM && s != NULL; 
       s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c), i++)
  {
    parameters[i] = atoi(s);
  }
  return 0;
}

CDI_CAN_EXPORT int initCanLineForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  int parameter[NUMBER_EXEC_PARAM];
  memset(parameter,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdiParseParamString(parameterList,parameter);
  return initCanLine(busLine,parameter,canMessage,cdiLine,numberDevices);
}

int initCanLine(int line, int *parameters, void (*callBack)(int, char *, int, int, int, int, int), int cdiLine, int numberDevices)
{ int i, k, ret, errorCode;
  CAN_LINE *pCan;
  DWORD dwCdiThreadId;
  struct timespec wait_time;
  int shared, value, retSem;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES )
  {
    cdilog("Open wrong Can line: %d; valid lines: 1 to %d\n",line, MAX_NUMBER_CAN_LINES);
    return -illegal_line;
  }
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd != 0 )
//#ifdef CDI_CAN_LINUX
//  if ( pCan->fd != 0 )
//#else
//  if ( pCan->fd != NULL )
//#endif
  {
    cdilog("Can line: %d already opened! \n",line+1);
    return 0;
  }

  pCan->canRate = transBaudRate( parameters[paramCanRate] );
  if ( pCan->canRate == -1 )
  {
    cdilog("Can line: %d; with Baudrate: %d not supported! \n",line,parameters[paramCanRate]);
    return -argument_list_error;
  }
  pCan->sdoNoBlockFlag = parameters[paramNoBlkSDO];
  pCan->sdoInterval = parameters[paramSdoInterval]; // 2013-8
  if ( pCan->sdoInterval < 0 ) pCan->sdoInterval = 0; //do not use sdo memory
  else if ( pCan->sdoInterval == 0 ) pCan->sdoInterval = 1000; //default 1000 ms 

  sprintf( pCan->devPath, "/dev/pcan%d", parameters[paramNodeNumber] );
  pCan->NumberDevices = numberDevices;
  pCan->cdiLine = cdiLine;
  pCan->receiveCanMessage = callBack;
  memset(&pCan->numRevMsg[0],0, sizeof(unsigned int) * numTypeMsg * MAX_CAN_IDS);

  pCan->txqueuesize = 8192; /*NTCAN_MAX_TX_QUEUESIZE; */
  pCan->rxqueuesize = 4096  /* NTCAN_MAX_RX_QUEUESIZE */;
  pCan->txtimeout = TXT_TIMEOUT;
  pCan->rxtimeout = 0;
  pCan->mode = 0;
#if defined(ESDCAN_H_)
  if ( parameters[paramNodeNumber] > ESD_LINE_OFFSET )
	pCan->net = parameters[paramNodeNumber] - ESD_LINE_OFFSET;
  else
	pCan->net = line;
  printf("ESD Can logical line modified: %d; Software line: %d\n",pCan->net, line+1);
  cdilog("ESD Can logical line modified: %d; Software line: %d\n",pCan->net, line+1);
#else
  pCan->net = line;
#endif
  ret = canDrvOpen(line, pCan->devPath, pCan->canRate, pCan->net, pCan->mode, pCan->txqueuesize, pCan->rxqueuesize, pCan->txtimeout, pCan->rxtimeout, &(pCan->fd), &errorCode);
  if ( ret == -1 )
    cdilog("canOpen() failed with error: %d; ESD can line: %d\n",errorCode, line + 1);
  else if ( ret == -2 )
    cdilog("canSetBaudrate() failed with error: %d; ESD can line: %d\n",errorCode, line + 1);
  else if ( ret == -3 )
      cdilog("canIdAdd() failed with error: %d; ESD can line: %d\n",errorCode, line + 1);
  else if ( ret == -10 )
      cdilog("LINUX_CAN_Open() failed with error: %d; Peak can line: %d\n",errorCode, line + 1);
  else if ( ret == -11)
    cdilog("CAN_Init() failed with error: %d; Peak can line: %d\n",errorCode, line + 1);

  if ( ret ) return -file_not_open;

  pCan->statusOK = -1;	/* mark as not init yet	*/

#ifdef CDI_CAN_LINUX
  if (pthread_create(&pCan->tIdMaster,NULL,GetCanDataThread,(void *)pCan) != 0)
#else
  if ( (pCan->tIdMaster = CreateThread(NULL,0,GetCanDataThread,(void *)pCan,0,&dwCdiThreadId)) == NULL )
#endif
  {
	  cdilog ("CreateThread foe CAN line: %d error",line + 1);
    terminateCanLine(line + 1);
	  return -un_allocated;
  }

  shared = 0;
  value = 1;
  if ((retSem=cdiGetSemaphore(&pCan->node_SemId,shared,value)) != 0)
  {
	  cdilog ("cdiGetSemaphore foe CAN line: %d error",line + 1);
    terminateCanLine(line + 1);
	  return -un_allocated;
  }

  /* Sleep(1000); *//* wait for driver init finished, ready for sending */
  wait_time.tv_sec = 1;
  wait_time.tv_nsec = 0;
  cdi_nanosleep( &wait_time, NULL );

  for (i=0; i<N_NODES; i++)
  {
//    cdiMutexInit(&pCan->sdoMutex[i], 1000 + i);
    pCan->mutexFlag[i] = 0;
    pCan->segmentFlag[i] = 0;
    pCan->sdoRequstTime[i] = 0;
    pCan->reqBufPointer[i] = NULL;
    pCan->asynIgnorCounter[i] = 0;

	for (k=0; k<numTypePDOs; k++)
	{
    cdiMutexInit(&pCan->pdoRevMsg[k][i].pdoMutex, 500 + i); // 2011-1-11
	  cdiMutexLock(&pCan->pdoRevMsg[k][i].pdoMutex);
	  pCan->pdoRevMsg[k][i].timeStemp = 0;
	  pCan->pdoRevMsg[k][i].dataLength = 0;
	  cdiMutexUnlock(&pCan->pdoRevMsg[k][i].pdoMutex);
	}
  }

  pCan->numSendErrors = 0;
  pCan->numSdoBuf = 0;  // 2013-8
  pCan->statusOK = 0;

  cdilog("canOpen() OK; can line: %d\n", line + 1);

  return (int)pCan->fd;
}

/****************************************/
/*	terminate can line					*/
/****************************************/
CDI_CAN_EXPORT int terminateCanLine(int line)
{ CAN_LINE *pCan;
  int ret = 0;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd == 0 ) return -file_not_open;
//#ifdef CDI_CAN_LINUX
//  if ( pCan->fd == 0 ) return -file_not_open;
//#else
//  if ( pCan->fd == NULL ) return -file_not_open;
//#endif

  if ( pCan->tIdMaster)
  {
    cdiCancelThread((int)pCan->tIdMaster);
    pCan->tIdMaster = 0;
  }
  ret = canDrvClose( &(canLine[line].fd) );
  if ( ret != 0 ) /*  NTCAN_SUCCESS */
  {
    cdilog("canClose() failed with error: %d; can line: %d\n",ret, line + 1);
    ret = -file_not_open;    
  }
  else cdilog("canClose() OK; can line: %d\n",line + 1);
#if defined(ESDCAN_H_) && defined(CDI_CAN_LINUX) && defined(ESDCAN_FIXED)
  ret = canDrvClose( &(canLine[line].readfd) );
  if ( ret != 0 ) /*  NTCAN_SUCCESS */
  {
    cdilog("read ESDCAN canClose() failed with error: %d; can line: %d\n",ret, line + 1);
    ret = -file_not_open;    
  }
  else cdilog("read ESDCAN canClose() OK; can line: %d\n",line + 1);
#endif
  return ret;
}
/****************************************/
/*	load Can card information			*/
/****************************************/
CDI_CAN_EXPORT int configCanCard(void)
{ int i;
  CAN_LINE *pCan;

  for ( i=0; i<MAX_NUMBER_CAN_LINES; i++ )
  {
    pCan = &(canLine[i]);
#ifdef JANZCAN_H_
    pCan->canRate = (i%2) == 0 ? BTR_500KB : BTR_20KB;
    pCan->canCardType = JANZ_PC_CARD;
#endif
    pCan->dpm_x = i/2;		// card number
    pCan->dpm_y = i%2;		// channel number
    pCan->lineNumber = i;
  }
  return 0;
}

