
/********************************************************/
/*	Janz CAN or other Can interface card for CDI				*/
/********************************************************/
#include "cdiCan.h" 	// Can card definition

int RegisterCanBus(void);
int canMsgSend(int line, short nodeId, short canOpenId, short length, char *data, short ObjNum, short subIndex, short wFlag, short sdoSegmentCmd);
int initCanLine(int line, int *parameters, void (*callBack)(int, char *, int, int, int, int, int), int cdiLine, int numberDevices);

CDI_CAN_EXPORT int filterCanbus(int *parameter, int accessFlag);
CDI_CAN_EXPORT void execCanbus(CdiRequestInfoBlk *pReq);
CDI_CAN_EXPORT int terminateCanLine(int line);
CDI_CAN_EXPORT int configCanCard(void);
CDI_CAN_EXPORT int initCanLineForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList);
CDI_CAN_EXPORT int canScanFcn(char *target,char *result,int resultBufferSize);
CDI_CAN_EXPORT int canTimeoutFcn(CdiRequestInfoBlk *pReq, int set, int taskIndex);
CDI_CAN_EXPORT int canDeviceReadyFcn(CdiRequestInfoBlk *pReq, int taskIndex);
CDI_CAN_EXPORT int canLoadRequest(CdiRequestInfoBlk *pReq, int taskIndex);

/* can card specific routines */
int canDrvOpen(int line, char *devPath, int canRate, int net, int mode, int txSize, int rxSize, int txTimeout, int rxTimeout, int *fd, int *erroCode);
int canDrvClose(int *fd);
int transBaudRate(int baudrate);
int readCanTelegram(int line, int fd, short *id, char *rtr, char *length, char *data);
int sendCanTelegram(int line, int fd, short sendId, char sendRTR, char sendLength, char *sendData, int *xmtStatus);

#ifdef JANZCAN_H_

#include "defs.h"	// Janz104 dpm headers
#include "vmod.h"
#include "icanif.h"
#include "calif.h"
#include "nmtm.h"
#include "dbtm.h"
#include "mcal.h"
#include "can_lib.h"
#include "dpm.h"
#include "mitop.h"

#endif /* JANZCAN_H_ */

#define MAX_NUMBER_CAN_LINES	4
#define N_NODES 128
#define numTypePDOs 5	//4 the last one is for nmt message 2016-2
#define MAX_SDO_BUFFER  1500
#define STORE_SDO_LENGTH  48    // in bytes

#define ESD_LINE_OFFSET	1000

/* open CAN parameter index */
enum openCanParamIndex
{
  paramCanRate,     /* e.g 500 in KHz */
  paramNodeNumber,  /* 32 for USB peak can  */
  paramNoBlkSDO,	  /* single SDO transfere without block, by setting to non zero value	*/
  paramSdoInterval,	/* using buffer for SDO reading, read intervel in ms	*/
};
enum execCanParamIndex
{
  msgType,   /* canMsgTypeIndex:   0, 1, 2, 3, 4, 5(any message) */
  canOpenObj, /* can open object number and pdo data offset*/
  canOpenSubIndex,
  canBroadcast, /* canID forced to be 0, one telegram for all devices */
  canAutoReceive, /* device sends data regulary (for Read or ReadthenWrite) */
  notused,
  canOpenId, /* can open id */
  canOpenReadId, /* can open id for read  */
  downloadFlag, /* current operation type for Block Transfer  */
  transState, /* current operation state for Block Transfer  */
  dataIndex, /* current read or written data index */
  dataOverflow, /* more data then receive buffer by sdo transfere */

  sdoMsgCopyTiStIndex = 13, /* index of sdo copy data time */
  sdoMsgTiStIndex = 14, /* index of sdo read timestamp */
  sdoMsgIndex = 15, /* index of sdo read message buffer starting from 1 */
};

enum canMsgTypeIndex
{
  sdoMsg,
  pdo1Msg,
  pdo2Msg,
  pdo3Msg,
  pdo4Msg,
  nmtMsg,
  emergencyMsg,
  notdefinedMsg,
  numTypeMsg
};

typedef union {			/* SDO Protocol	*/
  unsigned char value;
  struct {
    unsigned s:		1;	/* low byte for intel	*/
    unsigned e:		1;
    unsigned n:   2;
    unsigned x:   1;
    unsigned scs:	3;
  } s;
} SDO_PROTOCOL;

typedef union {			/* SDO Segment Protocol	*/
  unsigned char value;
  struct {
    unsigned c:		1;	/* low byte for intel	*/
    unsigned n:   3;
    unsigned t:   1;
    unsigned scs:	3;
  } s;
} SDO_SEG_PROTOCOL;

//--------------------------
//	Janz Card spesific
//--------------------------
#define NUM_CAN_TYPE		3	// cancard type:
#define JANZ_PC_CARD		1	// pccard-ican3-cal
#define JANZ_PC_CARD_RAW	2	// pccard-ican3-rawCan
#define PC104_CARD			3	// pc 104

#define FILTER_LOW	0
#define FILTER_HIGH	0x7ff

#define	HI_PRIORITY		1  // for CAN104 priority queues
#define	MID_PRIORITY	2
#define	LO_PRIORITY		3

typedef struct {			/* pdo message buffer	*/
  int timeStemp;
  short pdoType, dataLength;
  char	pdoData[8];
  cdi_mutex_t pdoMutex;
} PDO_MSG_BUF;

typedef struct {			/* sdo message buffer	*/
  int timeStemp;
  short id, index, subIndex, dataLength;
  char	sdoData[STORE_SDO_LENGTH];
} SDO_MSG_BUF;

typedef struct {			/* can line	*/
  int	cdiLine;	// cdi buffer index, start from 0,1,2

  int	fd;							// can file descriptor
#if defined(ESDCAN_H_) && defined(CDI_CAN_LINUX)
  int	readfd;				// can file descriptor
#endif
//#ifdef JANZCAN_H_
//  int	fd;							// can file descriptor
//#else
//  HANDLE	fd;							// can file descriptor
//#endif
  short	lineNumber, NumberDevices;	// can line number(>=0) and # of can modules
  char	devPath[20];				// driver path /dev/dpm_xy
  int	canRate;					// baudrate
  short	canCardType, sdoNoBlockFlag;	// PC card(CAL, RAW) or PC104
  short	dpm_x, dpm_y;				// modat number and can channels (0 or 1)

  int txqueuesize, rxqueuesize, txtimeout, rxtimeout;
  unsigned int mode, net;  // logical number

  short	statusOK;			// 1 initilized (at least one device alive)
  int	numSendErrors;		// number of can send errors within this line

  cdi_thread_t	tIdMaster;		// receive task variables
/*
#ifdef ESDCAN_H_
  CMSG cmsg[N_MSGS], *rcvPmsg, xmtMsg;
#endif
#ifdef PEAKCAN_H_
  TPCANMsg cmsg[N_MSGS], *rcvPmsg, xmtMsg;
#endif
  */
  unsigned int rcvStatus, recLength, recCnt, xmtLength;
  int xmtStatus;
  int segmentFlag[N_NODES], sdoRequstTime[N_NODES];  // segmentation transfere ON
  int asynIgnorCounter[N_NODES];		// counter
  CdiRequestInfoBlk *reqBufPointer[N_NODES];
  cdi_mutex_t sdoMutex[N_NODES]; /* for sdo transfere */
  char mutexFlag[N_NODES];

  void (*receiveCanMessage)(int cdiLine, char *data, int length, int retId, int rtr, int remoteNode, int remoteCommand);

  short	sendId;				// can message used in can send
  char	sendRTR, sendLength;
  char	sendData[8];

  short	retId;				// can message used in receive thread
  char	rtr, length;
  char	retData[8];
  short remoteNode, remoteCommand;

  unsigned int numRevMsg[numTypeMsg][MAX_CAN_IDS];

  PDO_MSG_BUF pdoRevMsg[numTypePDOs][MAX_CAN_IDS];

  short numSdoBuf, sdoInterval;
  SDO_MSG_BUF sdoRevMsg[MAX_SDO_BUFFER];

  sem_t node_SemId;
} CAN_LINE;
